#pylint: disable=no-member,too-many-lines,too-many-locals,too-many-statements
"""
Interface to access the model data in a *hynet* grid database.
"""

import logging

import numpy as np
import pandas as pd
from sqlalchemy import func
from sqlalchemy.orm import make_transient
from sqlalchemy.orm.exc import NoResultFound

from hynet.data.connection import DBTransaction
from hynet.data.structure import (DBInfo,
                                  DBScenario,
                                  DBBus,
                                  DBBranch,
                                  DBConverter,
                                  DBInjector,
                                  DBCapabilityRegion,
                                  DBSamplePoint,
                                  DBShunt,
                                  DBScenarioLoad,
                                  DBScenarioInjector,
                                  DBScenarioInactivity)
from hynet.scenario.capability import HalfSpace, CapRegion, ConverterCapRegion
from hynet.scenario.cost import PWLFunction
from hynet.scenario.representation import Scenario
from hynet.types_ import (hynet_id_,
                          hynet_float_,
                          hynet_complex_,
                          hynet_eps,
                          BusType,
                          BranchType,
                          InjectorType,
                          EntityType,
                          DBInfoKey)
from hynet.utilities.base import Timer

_log = logging.getLogger(__name__)


# === UTILITY FUNCTIONS =======================================================


def fix_hynet_id(series):
    """
    Fix the data type of the given nullable integer pandas series.

    When reading integer columns that contain NULL values using
    ``pandas.read_sql_table``, they are converted to float [1]_. This may lead
    to errors, e.g. the PWL function foreign keys cause key errors during
    indexing. This function fixes the type.

    References
    ----------
    .. [1] https://github.com/pandas-dev/pandas/issues/13049
    """
    if series.dtype == hynet_id_:
        return series
    series = series.astype(object)
    idx_nan = series.isnull()
    idx_val = ~idx_nan
    if np.any(np.mod(series.loc[idx_val], 1) != 0):
        raise ValueError("Some ID values are non-integer.")
    series.loc[idx_nan] = None
    series.loc[idx_val] = series.loc[idx_val].map(hynet_id_)
    return series


def get_max_bus_id(database):
    """
    Determine the highest bus ID in the specified *hynet* grid database.

    Parameters
    ----------
    database : DBConnection
        Connection to the *hynet* grid database.

    Returns
    -------
    result : .hynet_id_ or None
        The highest bus ID in the database or None if there are no buses.
    """
    with DBTransaction(database) as transaction:
        max_bus_id = transaction.query(
            func.max(DBBus.id).label("max_bus_id")).one().max_bus_id
    return max_bus_id


def get_max_scenario_id(database):
    """
    Determine the highest scenario ID in the specified *hynet* grid database.

    Parameters
    ----------
    database : DBConnection
        Connection to the *hynet* grid database.

    Returns
    -------
    result : .hynet_id_ or None
        The highest scenario ID in the database or None if there are no
        scenarios.
    """
    with DBTransaction(database) as transaction:
        max_scenario_id = transaction.query(
            func.max(DBScenario.id).label("max_scenario_id")).one().max_scenario_id
    return max_scenario_id


# === LOADING OF SCENARIO INFORMATION =========================================


def get_scenario_info(database):
    """
    Load a list of all scenarios from the specified *hynet* grid database.

    Parameters
    ----------
    database : DBConnection
        Connection to the *hynet* grid database.

    Returns
    -------
    pandas.DataFrame
        Data frame with a list of all scenarios, indexed by the *scenario ID*,
        which comprises the following columns:

        ``name``: (``str``)
            Name of the scenario.
        ``time``: (``hynet_float_``)
            Relative time in hours of the scenario (w.r.t. the scenario group
            start time).
        ``annotation``: (``str``)
            Annotation string for supplementary information.

    Raises
    ------
    ValueError
        In case that the database is empty.
    """
    if database.empty:
        raise ValueError("The database is empty.")
    scenario_info = pd.read_sql_table(DBScenario.__tablename__,
                                      database.engine)
    scenario_info.drop('loss_price', axis='columns', inplace=True)
    scenario_info.set_index('id', inplace=True)

    # Move 'name'-column to the front
    columns = scenario_info.columns.tolist()
    columns.insert(0, columns.pop(columns.index('name')))
    return scenario_info.reindex(columns=columns)


# === LOADING OF SCENARIOS ====================================================


def load_scenario(database, scenario_id=0):
    """
    Load the specified scenario from the *hynet* grid database.

    Parameters
    ----------
    database : DBConnection
        Connection to the *hynet* grid database.
    scenario_id : .hynet_id_, optional
        Identifier of the scenario; default is ``0``.

    Returns
    -------
    scenario : hynet.scenario.representation.Scenario
        Scenario object with the loaded data.

    Raises
    ------
    ValueError
        In case that the database is empty or the scenario with the specified
        ID was not found.
    """
    timer = Timer()

    if database.empty:
        raise ValueError("The database is empty.")

    scenario = Scenario()
    scenario.id = scenario_id
    scenario.database_uri = database.database_uri

    scenario.grid_name = database.get_setting(DBInfoKey.GRID_NAME)
    scenario.base_mva = hynet_float_(database.get_setting(DBInfoKey.BASE_MVA))
    scenario.description = database.get_setting(DBInfoKey.DESCRIPTION)

    with DBTransaction(database) as transaction:
        # General grid information
        try:
            scenario_result = \
                transaction.query(DBScenario).filter(
                    DBScenario.id == scenario_id).one()
        except NoResultFound:
            raise ValueError("Scenario with ID {:d} was not found."
                             .format(scenario_id))

        scenario.name = scenario_result.name
        scenario.time = hynet_float_(scenario_result.time)
        scenario.loss_price = hynet_float_(scenario_result.loss_price)
        scenario.annotation = scenario_result.annotation

        # Load the bus data
        scenario.bus = pd.read_sql_table(DBBus.__tablename__,
                                         database.engine)
        scenario.bus.set_index('id', inplace=True)

        # Map the bus types to their enumeration value
        scenario.bus['type'] = \
            scenario.bus['type'].map(lambda t: BusType(t)).values

        # Fix data type issues with the zone column
        scenario.bus['zone'] = fix_hynet_id(scenario.bus['zone'])

        # Load the shunt data (excluding any inactive shunts)
        scenario.bus['y_tld'] = np.zeros(scenario.num_buses,
                                         dtype=hynet_complex_)

        shunts = pd.read_sql_table(DBShunt.__tablename__, database.engine)
        shunts.set_index('id', inplace=True)

        shunt_inactivity = transaction.query(DBScenarioInactivity)\
            .filter(DBScenarioInactivity.scenario_id == scenario_id)\
            .filter(DBScenarioInactivity.entity_type == EntityType.SHUNT)
        try:
            shunts.drop([entity.entity_id for entity in shunt_inactivity],
                        axis='index', inplace=True)
        except KeyError:
            raise ValueError("Shunt inactivity refers to a nonexistent shunt.")
        shunts['y_tld'] = (shunts['p'] + 1j * shunts['q']) / scenario.base_mva

        for bus_id, shunt_value in zip(shunts['bus_id'], shunts['y_tld']):
            scenario.bus.at[bus_id, 'y_tld'] += shunt_value

        # Load the load data
        scenario.bus['load'] = np.zeros(scenario.num_buses,
                                        dtype=hynet_complex_)
        load_results = \
            transaction.query(DBScenarioLoad).filter(
                DBScenarioLoad.scenario_id == scenario_id)
        for load in load_results:
            scenario.bus.at[load.bus_id, 'load'] += load.p + 1j * load.q

        # Load the branch data
        scenario.branch = pd.read_sql_table(DBBranch.__tablename__,
                                            database.engine)
        scenario.branch.set_index('id', inplace=True)
        scenario.branch.rename(columns={'src_id': 'src', 'dst_id': 'dst'},
                               inplace=True)

        # Map the branch types to their enumeration value
        scenario.branch['type'] = \
            scenario.branch['type'].map(lambda t: BranchType(t)).values

        # Calculate the series impedance and shunt admittances
        scenario.branch['z_bar'] = \
            scenario.branch['r'] + 1j * scenario.branch['x']
        scenario.branch['y_src'] = 1j * scenario.branch['b_src']
        scenario.branch['y_dst'] = 1j * scenario.branch['b_dst']
        scenario.branch.drop(['r', 'x', 'b_src', 'b_dst'],
                             axis='columns', inplace=True)

        # Calculate the complex-valued voltage ratios of the transformers
        scenario.branch['rho_src'] = scenario.branch['ratio_src'] * np.exp(
            1j * scenario.branch['phase_src'] * np.pi / 180)
        scenario.branch['rho_dst'] = scenario.branch['ratio_dst'] * np.exp(
            1j * scenario.branch['phase_dst'] * np.pi / 180)
        scenario.branch.drop(['ratio_src', 'phase_src',
                              'ratio_dst', 'phase_dst'],
                             axis='columns', inplace=True)

        # Load the converter data
        scenario.converter = pd.read_sql_table(DBConverter.__tablename__,
                                               database.engine)
        scenario.converter.set_index('id', inplace=True)
        scenario.converter.rename(columns={'src_id': 'src', 'dst_id': 'dst'},
                                  inplace=True)

        # Prepare the conversion of cap. region IDs to cap. region objects
        cap_region = pd.read_sql_table(DBCapabilityRegion.__tablename__,
                                       database.engine)
        cap_region.set_index('id', inplace=True)

        def get_half_space(offset, slope):
            if offset is None or slope is None or \
                    np.isnan(offset) or np.isnan(slope):
                return None
            return HalfSpace(offset, slope)

        def get_cap_region_param(key):
            col = DBCapabilityRegion.__table__.c
            return {'p_bnd': (cap_region.at[key, col.p_min.name],
                              cap_region.at[key, col.p_max.name]),
                    'q_bnd': (cap_region.at[key, col.q_min.name],
                              cap_region.at[key, col.q_max.name]),
                    'lt': get_half_space(cap_region.at[key, col.lt_ofs.name],
                                         cap_region.at[key, col.lt_slp.name]),
                    'rt': get_half_space(cap_region.at[key, col.rt_ofs.name],
                                         cap_region.at[key, col.rt_slp.name]),
                    'lb': get_half_space(cap_region.at[key, col.lb_ofs.name],
                                         cap_region.at[key, col.lb_slp.name]),
                    'rb': get_half_space(cap_region.at[key, col.rb_ofs.name],
                                         cap_region.at[key, col.rb_slp.name])}

        # Map the converter cap. region IDs to objects
        scenario.converter[['cap_src', 'cap_dst']] = \
            scenario.converter[['cap_src_id', 'cap_dst_id']].applymap(
                lambda key: ConverterCapRegion(**get_cap_region_param(key)))
        scenario.converter.drop(['cap_src_id', 'cap_dst_id'],
                                axis='columns', inplace=True)

        # Load the injector data
        scenario.injector = pd.read_sql_table(DBInjector.__tablename__,
                                              database.engine)
        scenario.injector.set_index('id', inplace=True)
        scenario.injector.rename(columns={'bus_id': 'bus'}, inplace=True)

        # Map the injector types to their enumeration value
        scenario.injector['type'] = \
            scenario.injector['type'].map(lambda t: InjectorType(t)).values

        # Map the injector cap. region IDs to objects
        scenario.injector['cap'] = scenario.injector['cap_id'].map(
            lambda key: CapRegion(**get_cap_region_param(key)))
        scenario.injector.drop('cap_id', axis='columns', inplace=True)

        # Fix data type issues with the PWL function foreign keys
        scenario.injector['cost_p_id'] = \
            fix_hynet_id(scenario.injector['cost_p_id'])
        scenario.injector['cost_q_id'] = \
            fix_hynet_id(scenario.injector['cost_q_id'])

        # Prepare the conversion of PWL function IDs to PWL function objects
        sample_point = pd.read_sql_table(DBSamplePoint.__tablename__,
                                         database.engine)
        sample_point.set_index('id', inplace=True)

        def get_pwl_function(key):
            if key is None:
                return None
            x = sample_point.at[key, DBSamplePoint.__table__.c.x.name]
            y = sample_point.at[key, DBSamplePoint.__table__.c.y.name]
            return PWLFunction((x, y))

        # Map the PWL function region IDs to objects
        scenario.injector['cost_p'] = \
            scenario.injector['cost_p_id'].map(get_pwl_function)
        scenario.injector['cost_q'] = \
            scenario.injector['cost_q_id'].map(get_pwl_function)
        scenario.injector.drop(['cost_p_id', 'cost_q_id'],
                               axis='columns', inplace=True)

        # Adapt the injectors according to the scenario data
        injector_adjustment = \
            transaction.query(DBScenarioInjector).filter(
                DBScenarioInjector.scenario_id == scenario_id)

        def scale_pwl_function(pwl_function, scaling):
            if pwl_function is not None:
                pwl_function.scale(scaling)

        for adjustment in injector_adjustment:
            id_ = adjustment.injector_id

            # Adjust the cost function scaling
            scale_pwl_function(scenario.injector.at[id_, 'cost_p'],
                               adjustment.cost_scaling)
            scale_pwl_function(scenario.injector.at[id_, 'cost_q'],
                               adjustment.cost_scaling)

            # Adjust the capability region box constraint
            cap = scenario.injector.at[id_, 'cap']
            cap.p_min = adjustment.p_min
            cap.p_max = adjustment.p_max
            cap.q_min = adjustment.q_min
            cap.q_max = adjustment.q_max

        # Eliminate inactive entities
        # (REMARK: Inactive shunts are excluded via the SQL query above)
        inactive_entities = \
            transaction.query(DBScenarioInactivity).filter(
                DBScenarioInactivity.scenario_id == scenario_id)
        removed_buses = []
        for entity in inactive_entities:
            if entity.entity_type == EntityType.BUS:
                removed_buses.append(entity.entity_id)
            elif entity.entity_type == EntityType.BRANCH:
                scenario.branch.drop(entity.entity_id, inplace=True)
            elif entity.entity_type == EntityType.CONVERTER:
                scenario.converter.drop(entity.entity_id, inplace=True)
            elif entity.entity_type == EntityType.INJECTOR:
                scenario.injector.drop(entity.entity_id, inplace=True)
            elif entity.entity_type != EntityType.SHUNT:
                raise ValueError("Invalid entity is referenced for inactivity "
                                 + "({:s}: {:d}).".format(entity.entity_type,
                                                          entity.entity_id))
        scenario.remove_buses(removed_buses)

    _log.debug("Loaded scenario '{:s}' (id={:d}) of '{:s}' ({:.3f} sec.)"
               .format(scenario.name, scenario.id,
                       scenario.grid_name, timer.total()))
    return scenario


# === SOME SUPPORTING FUNCTIONS FOR THE INTERFACE FUNCTIONS BELOW =============


def _select_scenario_inactivities(scenario_id, transaction):
    return transaction.query(DBScenarioInactivity)\
        .filter(DBScenarioInactivity.scenario_id == scenario_id).all()


def _select_scenario_injectors(scenario_id, transaction):
    return transaction.query(DBScenarioInjector)\
        .filter(DBScenarioInjector.scenario_id == scenario_id).all()


def _select_scenario_loads(scenario_id, transaction):
    return transaction.query(DBScenarioLoad)\
        .filter(DBScenarioLoad.scenario_id == scenario_id).all()


def _select_scenario(scenario_id, transaction):
    return transaction.query(DBScenario)\
        .filter(DBScenario.id == scenario_id).one()


def _nan_to_none(value):
    if value is None:
        return None
    return None if np.isnan(value) else value


def _get_bus_data(scenario):
    counter_shunt_id = 1
    for index, pd_bus in scenario.bus.iterrows():
        bus_id = pd_bus.name  # Series name is the index of the row
        bus = DBBus(id=bus_id,
                    type=pd_bus.type,
                    ref=pd_bus.ref,
                    base_kv=pd_bus.base_kv,
                    v_min=pd_bus.v_min,
                    v_max=pd_bus.v_max,
                    zone=_nan_to_none(pd_bus.zone),
                    annotation=pd_bus.annotation)
        if pd_bus.load != 0:
            load = DBScenarioLoad(scenario_id=scenario.id,
                                  bus_id=bus_id,
                                  p=pd_bus.load.real,
                                  q=pd_bus.load.imag)
        else:
            load = None
        if pd_bus.y_tld != 0:
            shunt = DBShunt(id=counter_shunt_id,
                            bus_id=bus_id,
                            p=pd_bus.y_tld.real * scenario.base_mva,
                            q=pd_bus.y_tld.imag * scenario.base_mva,
                            annotation='')
            counter_shunt_id += 1
        else:
            shunt = None
        yield bus, load, shunt


def _get_branch_data(scenario):
    for index, pd_branch in scenario.branch.iterrows():
        branch_id = pd_branch.name  # Series name is the index of the row
        if pd_branch.y_src.real != 0 or pd_branch.y_dst.real != 0:
            _log.warning("Nonzero shunt conductance of branch {:d} is ignored"
                         .format(branch_id))
        branch = DBBranch(id=branch_id,
                          type=pd_branch.type,
                          src_id=pd_branch.src,
                          dst_id=pd_branch.dst,
                          r=pd_branch.z_bar.real,
                          x=pd_branch.z_bar.imag,
                          b_src=pd_branch.y_src.imag,
                          b_dst=pd_branch.y_dst.imag,
                          ratio_src=np.abs(pd_branch.rho_src),
                          phase_src=np.angle(pd_branch.rho_src) * 180 / np.pi,
                          ratio_dst=np.abs(pd_branch.rho_dst),
                          phase_dst=np.angle(pd_branch.rho_dst) * 180 / np.pi,
                          length=_nan_to_none(pd_branch.length),
                          rating=_nan_to_none(pd_branch.rating),
                          angle_min=_nan_to_none(pd_branch.angle_min),
                          angle_max=_nan_to_none(pd_branch.angle_max),
                          drop_min=_nan_to_none(pd_branch.drop_min),
                          drop_max=_nan_to_none(pd_branch.drop_max),
                          annotation=pd_branch.annotation)
        yield branch


def _get_cap_region_data(scenario, pd_cap_region):
    def get_halfspace_offset(halfspace):
        return None if halfspace is None else halfspace.offset

    def get_halfspace_slope(halfspace):
        return None if halfspace is None else halfspace.slope

    scenario.counter_cap_region_id += 1
    return DBCapabilityRegion(id=scenario.counter_cap_region_id,
                              p_min=pd_cap_region.p_min,
                              p_max=pd_cap_region.p_max,
                              q_min=pd_cap_region.q_min,
                              q_max=pd_cap_region.q_max,
                              lt_ofs=get_halfspace_offset(pd_cap_region.lt),
                              lt_slp=get_halfspace_slope(pd_cap_region.lt),
                              rt_ofs=get_halfspace_offset(pd_cap_region.rt),
                              rt_slp=get_halfspace_slope(pd_cap_region.rt),
                              lb_ofs=get_halfspace_offset(pd_cap_region.lb),
                              lb_slp=get_halfspace_slope(pd_cap_region.lb),
                              rb_ofs=get_halfspace_offset(pd_cap_region.rb),
                              rb_slp=get_halfspace_slope(pd_cap_region.rb),
                              annotation='')


def _get_converter_data(scenario):
    for index, pd_converter in scenario.converter.iterrows():
        converter_id = pd_converter.name  # Series name is the index of the row
        cap_src = _get_cap_region_data(scenario, pd_converter.cap_src)
        cap_dst = _get_cap_region_data(scenario, pd_converter.cap_dst)
        converter = DBConverter(id=converter_id,
                                src_id=pd_converter.src,
                                dst_id=pd_converter.dst,
                                cap_src_id=cap_src.id,
                                cap_dst_id=cap_dst.id,
                                loss_fwd=pd_converter.loss_fwd,
                                loss_bwd=pd_converter.loss_bwd,
                                loss_fix=pd_converter.loss_fix,
                                annotation=pd_converter.annotation)
        yield converter, cap_src, cap_dst


def _get_injector_data(scenario):
    def get_sample_points(pwl_function, function_id):
        sample_points = []
        for x, y in zip(*pwl_function.samples):
            sample_points.append(DBSamplePoint(id=function_id, x=x, y=y))
        return sample_points

    counter_cost_id = 1
    for index, pd_injector in scenario.injector.iterrows():
        injector_id = pd_injector.name  # Series name is the index of the row
        cap_region = _get_cap_region_data(scenario, pd_injector.cap)
        sample_points_p_id = None
        sample_points_p = []
        sample_points_q_id = None
        sample_points_q = []
        if pd_injector.cost_p is not None:
            sample_points_p_id = counter_cost_id
            sample_points_p = get_sample_points(pd_injector.cost_p,
                                                sample_points_p_id)
            counter_cost_id += 1
        if pd_injector.cost_q is not None:
            sample_points_q_id = counter_cost_id
            sample_points_q = get_sample_points(pd_injector.cost_q,
                                                sample_points_q_id)
            counter_cost_id += 1
        injector = DBInjector(id=injector_id,
                              type=pd_injector.type,
                              bus_id=pd_injector.bus,
                              cap_id=cap_region.id,
                              cost_p_id=sample_points_p_id,
                              cost_q_id=sample_points_q_id,
                              cost_start=pd_injector.cost_start,
                              cost_stop=pd_injector.cost_stop,
                              ramp_up=_nan_to_none(pd_injector.ramp_up),
                              ramp_down=_nan_to_none(pd_injector.ramp_down),
                              min_up=_nan_to_none(pd_injector.min_up),
                              min_down=_nan_to_none(pd_injector.min_down),
                              energy_min=_nan_to_none(pd_injector.energy_min),
                              energy_max=_nan_to_none(pd_injector.energy_max),
                              annotation=pd_injector.annotation)

        yield injector, cap_region, sample_points_p + sample_points_q


# === SAVING OF SCENARIOS =====================================================


def initialize_database(database, scenario):
    """
    Initialize an empty *hynet* grid database with the data of the scenario.

    The database is populated with the grid infrastructure and scenario data of
    the given scenario object. In the given scenario object, the scenario ID
    is reset to ``0`` and the database URI is updated.

    Parameters
    ----------
    database : DBConnection
        Connection to the *hynet* grid database.
    scenario : hynet.scenario.representation.Scenario
        Scenario object with the grid infrastructure and scenario data.

    Raises
    ------
    ValueError
        In case the database is not empty or any kind of data integrity or
        validity violation is detected.
    """
    if not database.empty:
        raise ValueError("The specified database is not empty.")

    scenario.verify()
    if np.any(scenario.branch['y_src'].values.real != 0) or \
            np.any(scenario.branch['y_dst'].values.real != 0):
        raise ValueError("Branches with a nonzero shunt conductance are "
                         "not supported.")

    scenario.id = 0
    scenario.database_uri = database.database_uri

    # Helper variable for generating IDs
    scenario.counter_cap_region_id = 0

    # Write general settings, grid infrastructure data, and scenario data
    with DBTransaction(database) as transaction:
        transaction.add(DBInfo(key=DBInfoKey.GRID_NAME,
                               value=scenario.grid_name))
        transaction.add(DBInfo(key=DBInfoKey.BASE_MVA,
                               value=scenario.base_mva))
        transaction.add(DBInfo(key=DBInfoKey.DESCRIPTION,
                               value=scenario.description))

        transaction.add(DBScenario(id=scenario.id,
                                   time=scenario.time,
                                   name=scenario.name,
                                   loss_price=scenario.loss_price,
                                   annotation=scenario.annotation))

        for bus, load, shunt in _get_bus_data(scenario):
            transaction.add(bus)
            if load is not None:
                transaction.add(load)
            if shunt is not None:
                transaction.add(shunt)

        for branch in _get_branch_data(scenario):
            transaction.add(branch)

        for converter, cap_src, cap_dst in _get_converter_data(scenario):
            transaction.add(converter)
            transaction.add(cap_src)
            transaction.add(cap_dst)

        for injector, cap, sample_points in _get_injector_data(scenario):
            transaction.add(injector)
            transaction.add(cap)
            transaction.add_all(sample_points)

    # Delete helper variable for generating IDs
    del scenario.counter_cap_region_id


def save_scenario(database, scenario, auto_id=True):
    """
    Save the provided scenario to the specified *hynet* grid database.

    The scenario information in the provided scenario object is extracted and
    saved to the specified *hynet* grid database. Note that a scenario can only
    capture the following changes:

        * Changes of the load (``scenario.bus['load']``)

        * Selected changes of injectors:

            - Scaling of the cost function for active and reactive power
              (``scenario.injector[['cost_p', 'cost_q']]``). The scaling must
              be the same for both cost functions of an injector.

            - Changes of the box constraints of the capability region
              (``p_min``, ``p_max``, ``q_min``, and ``q_max`` of the objects
              in ``scenario.injector['cap']``). Note that changes of the
              parameters of the half-spaces are *not* supported.)

        * Inactivity of buses, branches, converters, injectors, and shunts.
          For example, to deactivate (decommit) an injector, remove (drop) the
          respective row in the ``injector`` data frame of the scenario.

    Any other changes are considered as a modification of the grid
    infrastructure and cannot be captured as a scenario.

    Parameters
    ----------
    database : DBConnection
        Connection to the *hynet* grid database.
    scenario : Scenario
        Scenario object that represents the new scenario.
    auto_id : bool, optional
        If True (default), the scenario ID is set to the lowest available
        scenario ID in the database. Otherwise, the scenario ID of the
        Scenario object is applied.

    Raises
    ------
    ValueError
        In case the database is empty or any kind of data integrity or validity
        violation is detected.

    See Also
    --------
    initialize_database
    """
    if database.empty:
        raise ValueError("The specified database is empty.")

    scenario.verify()

    # Helper variable for generating IDs
    scenario.counter_cap_region_id = 0

    buses = set()
    loads = []
    shunts = set()
    branches = set()
    converters = set()
    converters_dict = dict()
    cap_regions = []
    injectors = set()
    injectors_dict = dict()
    sample_points = []

    for bus, load, shunt in _get_bus_data(scenario):
        buses.add(bus)
        if load is not None:
            loads.append(load)
        if shunt is not None:
            shunts.add(shunt)

    for branch in _get_branch_data(scenario):
        branches.add(branch)

    for converter, cap_src, cap_dst in _get_converter_data(scenario):
        converters.add(converter)
        converters_dict[converter.id] = converter
        cap_regions.append(cap_src)
        cap_regions.append(cap_dst)

    for injector, cap, sample_points_ in _get_injector_data(scenario):
        injectors.add(injector)
        injectors_dict[injector.id] = injector
        cap_regions.append(cap)
        sample_points += sample_points_

    with DBTransaction(database) as transaction:
        db_buses = set(transaction.query(DBBus))
        db_branches = set(transaction.query(DBBranch))
        db_converters = set(transaction.query(DBConverter))
        db_injectors = set(transaction.query(DBInjector))
        db_shunts = set(transaction.query(DBShunt))
        db_sample_points = transaction.query(DBSamplePoint)

        # Verify that the grid infrastructure is equivalent
        #
        # REMARK: The following set operations utilize specialized equality
        # operators of the DB objects, see ``structure.py``. They consider that
        # the IDs of shunts, sample points, and capability regions may differ.
        # Furthermore, they consider that the cost functions allow for a
        # scaling and that the capability regions may exhibit different box
        # constraints.
        message_prefix = "Difference of the grid infrastructure detected: "

        if buses - db_buses:
            raise ValueError(message_prefix + "Buses differ.")
        if branches - db_branches:
            raise ValueError(message_prefix + "Branches differ.")
        if converters - db_converters:
            raise ValueError(message_prefix + "Converters differ.")
        if injectors - db_injectors:
            raise ValueError(message_prefix + "Injectors differ.")
        if shunts - db_shunts:
            raise ValueError(message_prefix + "Shunts differ.")

        # Check if the capability regions of the converters are equal
        def compare_cap_region(cap1, cap2):
            return (cap1 == cap2 and  # Check half-spaces
                    cap1.p_min == cap2.p_min and
                    cap1.p_max == cap2.p_max and
                    cap1.q_min == cap2.q_min and
                    cap1.q_max == cap2.q_max)

        for converter in db_converters:
            if converter in converters:
                converter_new = converters_dict[converter.id]
                cap_src_new = cap_regions[converter_new.cap_src_id - 1]
                cap_dst_new = cap_regions[converter_new.cap_dst_id - 1]
                if not (compare_cap_region(cap_src_new, converter.cap_src) and
                        compare_cap_region(cap_dst_new, converter.cap_dst)):
                    raise ValueError(message_prefix +
                                     "Converter capability regions differ.")

        def load_sample_points(injector, extract_func_id):
            sp_db = db_sample_points.filter(
                DBSamplePoint.id == extract_func_id(injector)).all()
            sp_new = [sp for sp in sample_points
                      if sp.id == extract_func_id(injectors_dict[injector.id])]
            if len(sp_new) != len(sp_db):
                raise ValueError("Number of cost function samples differs.")

            if len(sp_new) == 0:
                return None, None

            def points2tuples(sample_points):
                x = np.array([sp.x for sp in sample_points])
                y = np.array([sp.y for sp in sample_points])
                idx = np.argsort(x)
                return x[idx], y[idx]

            return points2tuples(sp_db), points2tuples(sp_new)

        def get_function_scaling(injector, extract_func_id):
            sp_db, sp_new = load_sample_points(injector, extract_func_id)
            if sp_new is None:
                return None

            if any(np.abs(sp_db[0] - sp_new[0]) > 10 * hynet_eps):
                raise ValueError("The x-samples of cost functions differ.")

            with np.errstate(divide='ignore', invalid='ignore'):
                scaling = np.true_divide(sp_new[1], sp_db[1])
            scaling = scaling[np.abs(sp_db[1]) > 1e-6]
            if not scaling.size:
                if any(np.abs(sp_new[1]) > 1e-6):
                    raise ValueError("Scaling of zero cost is invalid.")
                return None
            if any(np.abs(scaling - scaling[0]) > 10 * hynet_eps):
                raise ValueError("Some cost function samples are not scaled "
                                 "uniformly.")
            return scaling[0]

        for injector in db_injectors:
            if injector in injectors:
                cap_new = cap_regions[injectors_dict[injector.id].cap_id - 1]
                if cap_new != injector.cap:
                    raise ValueError("The half-spaces of some injector "
                                     "capability regions differ.")

                scaling_p = get_function_scaling(injector,
                                                 lambda x: x.cost_p_id)
                scaling_q = get_function_scaling(injector,
                                                 lambda x: x.cost_q_id)

                if scaling_p is None and scaling_q is None:
                    injector.cost_scaling = 1.0
                elif scaling_p is not None and scaling_q is None:
                    injector.cost_scaling = scaling_p
                elif scaling_p is None and scaling_q is not None:
                    injector.cost_scaling = scaling_q
                else:
                    if not np.isclose(scaling_p, scaling_q):
                        raise ValueError("Some active and reactive cost "
                                         "functions are not scaled jointly.")
                    injector.cost_scaling = scaling_p

        # Verification passed. Save the scenario to the database.
        if auto_id:
            scenario.id = get_max_scenario_id(database) + 1
        scenario.database_uri = database.database_uri

        transaction.add(DBScenario(id=scenario.id,
                                   time=scenario.time,
                                   name=scenario.name,
                                   loss_price=scenario.loss_price,
                                   annotation=scenario.annotation))

        # Save load information
        for load in loads:
            load.scenario_id = scenario.id
        transaction.add_all(loads)

        # Save injector information
        for index, pd_injector in scenario.injector.iterrows():
            db_injector = None
            for injector in db_injectors:
                if injector.id == pd_injector.name:
                    db_injector = injector
                    break
            assert db_injector is not None  # This is already checked above...

            # Skip injector reconfiguration if there was no modification
            if (np.isclose(db_injector.cost_scaling, 1.0) and
                    db_injector.cap.p_min == pd_injector.cap.p_min and
                    db_injector.cap.p_max == pd_injector.cap.p_max and
                    db_injector.cap.q_min == pd_injector.cap.q_min and
                    db_injector.cap.q_max == pd_injector.cap.q_max):
                continue

            transaction.add(
                DBScenarioInjector(scenario_id=scenario.id,
                                   injector_id=pd_injector.name,
                                   p_min=pd_injector.cap.p_min,
                                   p_max=pd_injector.cap.p_max,
                                   q_min=pd_injector.cap.q_min,
                                   q_max=pd_injector.cap.q_max,
                                   cost_scaling=db_injector.cost_scaling))

        # Save inactivity information
        inactive_buses = db_buses - buses
        inactive_branches = db_branches - branches
        inactive_converters = db_converters - converters
        inactive_shunts = db_shunts - shunts
        inactive_injectors = db_injectors - injectors

        def create_inactivity_datasets(entity_type, inactive_entities):
            return map(lambda entity:
                       DBScenarioInactivity(scenario_id=scenario.id,
                                            entity_type=entity_type,
                                            entity_id=entity.id),
                       inactive_entities)

        transaction.add_all(create_inactivity_datasets(EntityType.BUS,
                                                       inactive_buses))
        transaction.add_all(create_inactivity_datasets(EntityType.BRANCH,
                                                       inactive_branches))
        transaction.add_all(create_inactivity_datasets(EntityType.CONVERTER,
                                                       inactive_converters))
        transaction.add_all(create_inactivity_datasets(EntityType.SHUNT,
                                                       inactive_shunts))
        transaction.add_all(create_inactivity_datasets(EntityType.INJECTOR,
                                                       inactive_injectors))

    # Delete helper variable for generating IDs
    del scenario.counter_cap_region_id


# === REMOVING OF SCENARIOS ===================================================


def remove_scenarios(database, scenario_ids):
    """
    Remove the specified scenarios from the *hynet* grid database file.

    Parameters
    ----------
    database : DBConnection
        Connection to the *hynet* grid database.
    scenario_ids : list[.hynet_id_]
        List of identifiers for the scenarios that shall be removed.

    Raises
    ------
    ValueError
        In case that a scenario with the specified ID was not found.
    """
    # Check if all scenarios exist
    scenario_info = get_scenario_info(database)
    for scenario_id in scenario_ids:
        if not scenario_info.index.contains(scenario_id):
            raise ValueError("The database does not contain a scenario with "
                             "the ID '{:d}'.".format(scenario_id))

    # Remove the scenarios
    for scenario_id in scenario_ids:
        with DBTransaction(database) as transaction:
            scenario = \
                _select_scenario(scenario_id, transaction)
            scenarios_loads = \
                _select_scenario_loads(scenario_id, transaction)
            scenario_injectors = \
                _select_scenario_injectors(scenario_id, transaction)
            scenario_inactivities = \
                _select_scenario_inactivities(scenario_id, transaction)

            transaction.delete_all(scenarios_loads)
            transaction.delete_all(scenario_injectors)
            transaction.delete_all(scenario_inactivities)
            transaction.delete(scenario)


# === COPYING OF SCENARIOS ====================================================


def copy_scenarios(source_database, destination_database,
                   scenario_ids=None, bus_id_map=None):
    """
    Copy the scenarios from the source to the destination *hynet* grid database.

    **Caution:** Note that this function does *not* verify that the grid
    infrastructure of the source and destination database is compatible, it
    only copies the content of the scenario information tables for the
    requested scenarios. In case a scenario with a specified ID already exists
    in the destination database, it is *skipped*.

    Parameters
    ----------
    source_database : DBConnection
        Connection to the source *hynet* grid database.
    destination_database : DBConnection
        Connection to the destination *hynet* grid database.
    scenario_ids : list[.hynet_id_], optional
        List of scenario identifiers that shall be copied. By default, all
        scenarios except the "base case" with scenario ID 0 are copied.
    bus_id_map : pandas.Series, optional
        If provided, this series, *indexed by the bus IDs in the source
        database*, specifies the mapping of a bus ID in the source database to
        a bus ID in the destination database. This is necessary, e.g., after
        a network reduction, where certain buses are combined and the load
        shall be accumulated at the aggregated bus. By default, a one-to-one
        mapping is considered.
    """
    if scenario_ids is None:
        scenario_ids = get_scenario_info(source_database).index

    dst_scenario_ids = get_scenario_info(destination_database).index

    if bus_id_map is None:
        _id_range = range(get_max_bus_id(source_database) + 1)
        bus_id_map = pd.Series(_id_range, index=_id_range)

    for scenario_id in scenario_ids:
        if dst_scenario_ids.contains(scenario_id):
            _log.warning("Scenario with ID '{:d}' already exists. Skipping it."
                         .format(scenario_id))
            continue

        with DBTransaction(source_database) as src_transaction, \
                DBTransaction(destination_database) as dst_transaction:
            # Load scenario information from the source database
            scenario = \
                _select_scenario(scenario_id, src_transaction)
            loads = \
                _select_scenario_loads(scenario_id, src_transaction)
            injectors = \
                _select_scenario_injectors(scenario_id, src_transaction)
            inactivities = \
                _select_scenario_inactivities(scenario_id, src_transaction)

            # Prepare data for copying
            make_transient(scenario)

            new_loads = dict()
            for load in loads:
                make_transient(load)
                load.bus_id = int(bus_id_map.at[load.bus_id])

                # Accumulate the load for every bus
                if load.bus_id in new_loads:
                    new_loads[load.bus_id].p += load.p
                    new_loads[load.bus_id].q += load.q
                else:
                    new_loads[load.bus_id] = load

            for injector in injectors:
                make_transient(injector)

            for inactivity in inactivities:
                make_transient(inactivity)
                if inactivity.entity_type == EntityType.BUS:
                    inactivity.entity_id = \
                        int(bus_id_map.at[inactivity.entity_id])

            # Write scenario information to the destination database
            dst_transaction.add(scenario)
            dst_transaction.add_all(new_loads.values())
            dst_transaction.add_all(injectors)
            dst_transaction.add_all(inactivities)



"""
Management of worker processes in *hynet*.

Parameters
----------
workers : WorkerManager
    Worker manager created at import time for parallel processing within the
    *hynet* package.
"""

import logging
from os import cpu_count
import multiprocessing
import numbers
import functools
from itertools import starmap

from hynet.utilities.base import Timer
import hynet.config as config

_log = logging.getLogger(__name__)


def pool_operation(method):
    """
    Decorates a worker operation method with a pool state verification.

    The worker manager maintains its pool of workers lazily, i.e., it is only
    created on demand. Due to this, every potential pool operation in the
    worker manager class has to verify and, if necessary, update the pool
    state prior to the operation itself. This decorator manages this task. In
    the worker operation method, the code only needs to adapt to the
    availability of the pool.
    """
    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        self._check_pool()
        return method(self, *args, **kwargs)
    return wrapper


class WorkerManager:
    """
    Manage operations on a pool of workers.

    This class provides operations on a pool of worker processes in case that
    parallel processing in *hynet* is enabled and, otherwise, it performs the
    operations in the current process. The pool of workers is maintained
    lazily, i.e., it is created on demand.

    See Also
    --------
    hynet.config
    """
    def __init__(self):
        self._pool = None
        self._num_workers = cpu_count()
        if self._num_workers is None:
            self._num_workers = 1

    @property
    def num_workers(self):
        """
        Return the number of worker processes.

        **These workers are only active if parallel processing is enabled.**

        See Also
        --------
        hynet.config
        """
        return self._num_workers

    @num_workers.setter
    def num_workers(self, value):
        """
        Set the number of worker processes.

        **These workers are only active if parallel processing is enabled.** If
        a pool with a different number of worker processes is already open, it
        is closed immediately and reopened on demand with the updated number of
        workers.

        See Also
        --------
        hynet.config
        """
        if not isinstance(value, numbers.Integral) or value < 1:
            raise ValueError("The number of workers must be a positive integer.")
        if value == self._num_workers:
            return
        self._num_workers = value
        self.close_pool()

    def _open_pool(self):
        """Open a pool of worker processes."""
        if self._pool is not None:
            raise RuntimeError("There is already an open pool of workers.")
        if self._num_workers == 1:
            _log.debug("Skipping pool creation due to use of a single worker.")
            return
        timer = Timer()
        self._pool = multiprocessing.Pool(processes=self._num_workers)
        _log.debug("Opened pool with {:d} workers ({:.3f} sec.)"
                   .format(self._num_workers, timer.total()))

    def close_pool(self):
        """Close the pool of worker processes."""
        if self._pool is None:
            return
        timer = Timer()
        self._pool.close()
        self._pool.join()
        self._pool = None
        _log.debug("Closed worker pool ({:.3f} sec.)".format(timer.total()))

    def _check_pool(self):
        """
        Verify and, if necessary, update the state of the worker pool.

        See Also
        --------
        pool_operation: Decorator for a worker operation method.
        """
        if config.GENERAL['parallelize']:
            if self._pool is None:
                self._open_pool()
        else:
            if self._pool is not None:
                _log.debug("Parallel processing was disabled. Closing the pool.")
            self.close_pool()

    @property
    def is_multiprocessing(self):
        """Return True if the operations utilize multiprocessing."""
        self._check_pool()
        return self._pool is not None

    @pool_operation
    def map(self, func, iterable, **kwds):
        """
        Apply the function to every item (single argument) in the iterable.

        Parameters
        ----------
        func : function
            Function that shall be applied to the items in the iterable.
        iterable : iterable
            Iterable containing the function arguments. An item in this
            iterable is the single argument passed to the function.
        kwds : dict, optional
            If parallel processing is enabled, the keyword arguments are passed
            to the ``map``-method of ``multiprocessing.Pool``.

        Returns
        -------
        result : list
            List of the function result for every item in the iterable.

        See Also
        --------
        multiprocessing.Pool.map
        """
        if self._pool is None:
            return list(map(func, iterable))
        return self._pool.map(func, iterable, **kwds)

    @pool_operation
    def starmap(self, func, iterable, **kwds):
        """
        Apply the function to every item (argument tuple) in the iterable.

        Parameters
        ----------
        func : function
            Function that shall be applied to the items in the iterable.
        iterable : iterable
            Iterable containing the function arguments. An item in this
            iterable is a tuple of the arguments passed to the function.
        kwds : dict, optional
            If parallel processing is enabled, the keyword arguments are passed
            to the ``starmap``-method of ``multiprocessing.Pool``.

        Returns
        -------
        result : list
            List of the function result for every item in the iterable.

        See Also
        --------
        multiprocessing.Pool.starmap
        """
        if self._pool is None:
            return list(starmap(func, iterable))
        return self._pool.starmap(func, iterable, **kwds)


workers = WorkerManager()   # One instance for use within the *hynet* package.



"""
General utilities.
"""

import logging
from timeit import default_timer

import numpy as np
from scipy.sparse import diags

from hynet.types_ import hynet_int_, hynet_float_, hynet_complex_, hynet_sparse_

_log = logging.getLogger(__name__)


def create_sparse_matrix(i, j, data, m, n, dtype=hynet_complex_):
    """
    Return an ``m``-by-``n`` sparse matrix ``A`` with ``A[i[k], j[k]] = data[k]``.
    """
    return hynet_sparse_((data, (i, j)), shape=(m, n), dtype=dtype)


def create_sparse_zero_matrix(m, n):
    """Return an ``m``-by-``n`` all-zeros matrix."""
    return hynet_sparse_((m, n), dtype=hynet_float_)


def create_sparse_diag_matrix(diagonal, dtype=hynet_complex_):
    """Return a diagonal matrix."""
    return hynet_sparse_(diags(diagonal), dtype=dtype)


def create_dense_vector(i, data, n, dtype=hynet_complex_):
    """Create ``n``-dim. vector ``x`` with ``x[i] = data[i]``."""
    # This is somewhat hacky... May be revised if performance is bad ;)
    return create_sparse_matrix(np.zeros(len(i)), i, data, 1, n,
                                dtype=dtype).toarray()[0]


def truncate_with_ellipsis(string, max_len):
    """
    Truncate the string with an ellipsis if its length exceeds ``max_len``.
    """
    return (string[:max_len-3] + '...') if len(string) > max_len else string


def partition_iterable(iterable, num_blocks):
    """
    Partition the iterable into the specified number of consecutive blocks.

    Parameters
    ----------
    iterable : Iterable
        Iterable to be partitioned.
    num_blocks : .hynet_int_
        Number of blocks into which the iterable shall be partitioned.

    Yields
    ------
    blocks : Iterable
        Blocks of the iterable.
    """
    num_elements = len(iterable)
    if num_elements < 1:
        raise ValueError("The iterable is empty.")
    if num_blocks < 1:
        raise ValueError("The number of blocks must be strictly positive.")
    if num_blocks > num_elements:
        raise ValueError("The number of blocks exceeds the iterable's length.")

    block_size = hynet_int_(np.ceil(num_elements / num_blocks))
    if block_size * (num_blocks - 1) >= num_elements:
        block_size -= 1

    partition = block_size * np.arange(num_blocks, dtype=hynet_int_)
    i = 1
    while num_elements - partition[-1] > block_size:
        partition[i:] += 1
        i += 1

    for i in range(num_blocks - 1):
        yield iterable[partition[i]:partition[i + 1]]
    yield iterable[partition[-1]:]


def validate_mosek_license():
    """Return True if the MOSEK installation has a valid license."""
    try:
        import mosek
    except ImportError:
        return False
    try:
        with mosek.Env() as env:
            with env.Task() as task:
                task.optimize()
    except mosek.Error as exception:
        _log.info("MOSEK is not properly installed: " + str(exception))
        return False
    return True


class Timer:
    """
    Measure time since ``Timer`` object creation or time measurements.
    """
    def __init__(self):
        self._start_time = self._last_time = default_timer()

    def interval(self):
        """
        Return the time in seconds since the last measurement or object creation.

        For the first measurement, the time elapsed since object creation is
        returned. From the second measurement onwards, the time elapsed since
        the last measurement is returned.
        """
        current_time = default_timer()
        time_elapsed = current_time - self._last_time
        self._last_time = current_time
        return time_elapsed

    def total(self):
        """Return the time in seconds since object creation."""
        return default_timer() - self._start_time

    def reset(self):
        """Reset the timer."""
        self._start_time = self._last_time = default_timer()



"""
MOSEK-based solver for the *hynet*-specific QCQP problem (SD and SOC relaxation).
"""

import logging
import sys
import abc

import numpy as np
import mosek

from hynet.types_ import hynet_float_, SolverType, SolverStatus
from hynet.utilities.base import Timer
from hynet.qcqp.solver import SolverInterface
from hynet.qcqp.problem import QCQPPoint
from hynet.qcqp.result import QCQPResult

_log = logging.getLogger(__name__)


def parse_status(status):
    """Parse the MOSEK status into a ``SolverStatus``."""
    if status == mosek.solsta.optimal:
        return SolverStatus.SOLVED
    if status == mosek.solsta.near_optimal:
        return SolverStatus.INACCURATE
    if status in [mosek.solsta.prim_infeas_cer,
                  mosek.solsta.dual_infeas_cer,
                  mosek.solsta.near_prim_infeas_cer,
                  mosek.solsta.near_dual_infeas_cer]:
        return SolverStatus.INFEASIBLE
    return SolverStatus.FAILED


class SolverBase(SolverInterface):
    """Base class for MOSEK-based solvers for a relaxation of the QCQP."""

    # REMARK concerning the logger callback:
    #   Originally, the class accepted an argument ``log_stream`` of type
    #   ``io.TextIOBase`` for the object initialization, which was set to
    #   ``sys.stdout`` by default and used as the output stream for the logging
    #   information in verbose mode. However, as a consequence of the solver
    #   object containing an open stream as an attribute, it was rendered
    #   non-serializable with pickle. As a customization of the logging output
    #   stream is typically not needed, this implementation hard-wires the
    #   logging to ``sys.stdout`` to render the solver objects 'pickleable'.
    def _logger(self, message):
        """Logger callback for MOSEK"""
        sys.stdout.write(message)
        sys.stdout.flush()

    @staticmethod
    @abc.abstractmethod
    def _add_voltage_matrix_constraints(task, qcqp, dim_x, ofs, x_lb, x_ub, x_bnd):
        """Add the SOC or PSD constraint(s) on the ``v_tld``-part of ``x``."""

    def solve(self, qcqp):
        """
        Solve the given QCQP via a relaxation using MOSEK.

        The relaxation-specific constraints on the bus voltage matrix are
        added via the (abstract) method ``_add_voltage_matrix_constraints``,
        which must be implemented in a derived class.

        Parameters
        ----------
        qcqp : QCQP
            Specification of the quadratically-constrained quadratic problem.

        Returns
        -------
        result : QCQPResult
            Solution of the relaxed problem associated with the QCQP.
        """
        # We solve the following problem:
        #
        #  min   c^T x   s.t.   Ax == b,  Bx <= d,  x_lb <= x <= x_ub
        #   x                ...and SOC or PSD constraint(s) on v_tld-part of x
        timer = Timer()
        (c, A, b, B, d, x_lb, x_ub) = qcqp.get_vectorization()
        dim_x = c.shape[0]

        # Convert from the COO to the CSR sparse matrix format: In the COO
        # format, there can be duplicate entries (i.e., more than one value for
        # the same element in the matrix, which are to be added), but this is
        # not supported by MOSEK's ``putaijlist``. To this end, the matrices
        # are converted to the CSR format, which sums up the duplicate entries.
        c = c.tocsr()
        A = A.tocsr()
        B = B.tocsr()

        x_lb_is_nan = np.isnan(x_lb)
        x_ub_is_nan = np.isnan(x_ub)
        x_bnd = np.array([mosek.boundkey.ra] * dim_x)
        x_bnd[np.logical_and(x_lb_is_nan, x_ub_is_nan)] = mosek.boundkey.fr
        x_bnd[np.logical_and(~x_lb_is_nan, x_ub_is_nan)] = mosek.boundkey.lo
        x_bnd[np.logical_and(x_lb_is_nan, ~x_ub_is_nan)] = mosek.boundkey.up
        x_bnd[x_lb == x_ub] = mosek.boundkey.fx

        # For the following code, I used the MOSEK documentation provided at
        #  1) http://docs.mosek.com/8.1/pythonapi.pdf
        #  2) https://docs.mosek.com/8.1/pythonapi/optimizer-task.html
        with mosek.Env() as env:
            if self.verbose:
                env.set_Stream(mosek.streamtype.log, self._logger)

            with env.Task() as task:
                if self.verbose:
                    task.set_Stream(mosek.streamtype.log, self._logger)

                # Variable definition and bounds
                task.appendvars(dim_x)
                task.putvarboundlist(range(dim_x), x_bnd, x_lb, x_ub)

                # Objective
                task.putobjsense(mosek.objsense.minimize)
                idx_nz = c.nonzero()[0]
                task.putclist(idx_nz, c[idx_nz].toarray()[:, 0])

                # Equality constraints
                ofs = 0
                N = b.shape[0]
                task.appendcons(N)
                idx_nz = A.nonzero()
                task.putaijlist(ofs + idx_nz[0], idx_nz[1],
                                np.asarray(A[idx_nz[0], idx_nz[1]])[0])
                task.putconboundlist(range(ofs, ofs + N),
                                     [mosek.boundkey.fx] * N, b, b)
                ofs += N

                # Inequality constraints
                M = d.shape[0]
                task.appendcons(M)
                idx_nz = B.nonzero()
                task.putaijlist(ofs + idx_nz[0], idx_nz[1],
                                np.asarray(B[idx_nz[0], idx_nz[1]])[0])
                task.putconboundlist(range(ofs, ofs + M),
                                     [mosek.boundkey.up] * M, np.zeros(M), d)
                ofs += M

                # Add relaxation-specific constraints on the voltage matrix
                # (This method is overridden appropriately in derived classes)
                self._add_voltage_matrix_constraints(task, qcqp, dim_x, ofs,
                                                     x_lb, x_ub, x_bnd)

                # Set solver parameters
                for key, value in self.param.items():
                    task.putparam(key, str(value))

                # Optimization and solution recovery
                _log.debug(self.type.name + " ~ Modeling ({:.3f} sec.)"
                           .format(timer.interval()))
                task.optimize()
                solver_time = timer.interval()
                _log.debug(self.type.name + " ~ Solver ({:.3f} sec.)"
                           .format(solver_time))

                if self.verbose:
                    task.solutionsummary(mosek.streamtype.log)

                status = parse_status(task.getsolsta(mosek.soltype.itr))

                if status in [SolverStatus.SOLVED, SolverStatus.INACCURATE]:
                    x = np.ndarray(dim_x, hynet_float_)
                    task.getxxslice(mosek.soltype.itr, 0, dim_x, x)
                    (V, f, s, z) = qcqp.split_vectorization_optimizer(x)
                    v = self.rank1approx(V, qcqp.edges, qcqp.roots)
                    optimizer = QCQPPoint(v, f, s, z)

                    dv_lb = np.ndarray(dim_x, dtype=hynet_float_)
                    task.getslxslice(mosek.soltype.itr, 0, dim_x, dv_lb)
                    dv_lb[np.isnan(x_lb)] = np.nan
                    dv_lb = qcqp.split_vectorization_bound_dual(dv_lb)

                    dv_ub = np.ndarray(dim_x, dtype=hynet_float_)
                    task.getsuxslice(mosek.soltype.itr, 0, dim_x, dv_ub)
                    dv_ub[np.isnan(x_ub)] = np.nan
                    dv_ub = qcqp.split_vectorization_bound_dual(dv_ub)

                    dv_eq_lb = np.ndarray(N, dtype=hynet_float_)
                    task.getslcslice(mosek.soltype.itr, 0, N, dv_eq_lb)
                    dv_eq_ub = np.ndarray(N, dtype=hynet_float_)
                    task.getsucslice(mosek.soltype.itr, 0, N, dv_eq_ub)
                    dv_eq = dv_eq_ub - dv_eq_lb

                    dv_ineq = np.ndarray(M, dtype=hynet_float_)
                    task.getsucslice(mosek.soltype.itr, N, N + M, dv_ineq)

                    optimal_value = task.getprimalobj(mosek.soltype.itr)

                    result = QCQPResult(qcqp, self, status, solver_time,
                                        optimizer=optimizer,
                                        V=V,
                                        optimal_value=optimal_value,
                                        dv_lb=dv_lb,
                                        dv_ub=dv_ub,
                                        dv_eq=dv_eq,
                                        dv_ineq=dv_ineq)
                else:
                    result = QCQPResult(qcqp, self, status, solver_time)

        _log.debug(self.type.name + " ~ Result creation ({:.3f} sec.)"
                   .format(timer.interval()))
        _log.debug(self.type.name + " ~ Total time for solver ({:.3f} sec.)"
                   .format(timer.total()))
        return result


class SOCRSolver(SolverBase):
    """
    MOSEK-based solver for a second-order cone relaxation of the QCQP.

    For information on solver-specific parameters, see e.g. [1]_, [2]_.

    See Also
    --------
    hynet.qcqp.solver.SolverInterface : Abstract base class for QCQP solvers.

    References
    ----------
    .. [1] https://docs.mosek.com/8.1/opt-server/param-groups.html
    .. [2] https://docs.mosek.com/8.1/pythonapi/solving-conic.html
    """

    @property
    def name(self):
        return "MOSEK"

    @property
    def type(self):
        return SolverType.SOCR

    def solve(self, qcqp):  # pylint: disable=useless-super-delegation
        """
        Solve the given QCQP via its second-order cone relaxation using MOSEK.

        Parameters
        ----------
        qcqp : QCQP
            Specification of the quadratically-constrained quadratic problem.

        Returns
        -------
        result : QCQPResult
            Solution of the SOCR associated with the QCQP.
        """
        # We solve the following problem:
        #
        #  min   c^T x   s.t.   Ax == b,  Bx <= d,  x_lb <= x <= x_ub
        #   x                   ...and SOC constraints on v_tld-part of x
        return super().solve(qcqp)

    @staticmethod
    def _add_voltage_matrix_constraints(task, qcqp, dim_x, ofs, x_lb, x_ub, x_bnd):
        """Add the SOC constraints on the ``v_tld``-part of ``x``."""
        # Second-order cone constraints on V (cf. QCQP.get_vectorization)
        #
        # PSD constraint on 2x2 principal submatrix =>
        #  (a) diagonal is nonnegative -> ensured by LB on voltage mag.
        #  (b) determinant is nonnegative -> rotated SOC:
        #
        #               V_ij * conj(V_ij) <= V_ii * V_jj
        #                             <=>
        #          real(V_ij)^2 + imag(V_ij)^2 <= V_ii * V_jj
        #
        #      where i is the source bus and j is the destination bus
        #      of the considered edge. Note that MOSEK requires
        #      rotated second-order cones to be specified as
        #
        #              2 * x_0 * x_1 >= x_2^2 + x_3^2
        #
        #      i.e., there is an additional scaling factor of 2.
        #
        # MOSEK limits variables to being a member of only one cone. If
        # multiple edges connect to a bus, the corresponding diagonal
        # element is a member of more than one cone. Due to this, the
        # diagonal elements V_ii and V_jj are duplicated for every
        # second-order cone, i.e., for every edge k with source node i
        # and destination node j an additional variable
        # w_k = [V_ii/2, V_jj]^T is introduced. Thus, the augmented
        # optimization variable x' is
        #
        #            x' = [x^T, w_0^T, ..., w_(K-1)^T]^T
        #
        # where K is the number of edges.

        # Add the variables w_k. Note that the diagonal elements of V
        # are at the "beginning" of x and the bounds can be reused.
        (e_src, e_dst) = qcqp.edges
        K = qcqp.num_edges
        task.appendvars(2*K)
        idx_w_src = range(dim_x, dim_x + 2*K, 2)
        idx_w_dst = range(dim_x + 1, dim_x + 2*K, 2)
        task.putvarboundlist(idx_w_src, x_bnd[e_src],
                             x_lb[e_src]/2, x_ub[e_src]/2)
        task.putvarboundlist(idx_w_dst, x_bnd[e_dst],
                             x_lb[e_dst], x_ub[e_dst])

        task.appendcons(2*K)
        idx_crt_w_src = range(ofs, ofs + 2*K, 2)
        idx_crt_w_dst = range(ofs + 1, ofs + 2*K, 2)
        task.putaijlist(idx_crt_w_src, e_src, 0.5 * np.ones(K))
        task.putaijlist(idx_crt_w_src, idx_w_src, -np.ones(K))
        task.putaijlist(idx_crt_w_dst, e_dst, np.ones(K))
        task.putaijlist(idx_crt_w_dst, idx_w_dst, -np.ones(K))
        task.putconboundlist(range(ofs, ofs + 2*K),
                             [mosek.boundkey.fx] * 2*K,
                             np.zeros(2*K), np.zeros(2*K))
        ofs += 2*K

        for k in range(K):
            task.appendcone(mosek.conetype.rquad, 0,
                            [idx_w_src[k], idx_w_dst[k],
                             qcqp.dim_v + k, qcqp.dim_v + K + k])


class SDRSolver(SolverBase):
    """
    MOSEK-based solver for a semidefinite relaxation of the QCQP.

    For information on solver-specific parameters, see e.g. [1]_. With the
    default setting, the dual variables turned out to be inaccurate, due to
    which the default tolerance on dual feasibility is adapted.

    See Also
    --------
    hynet.qcqp.solver.SolverInterface : Abstract base class for QCQP solvers.

    References
    ----------
    .. [1] https://docs.mosek.com/8.1/opt-server/param-groups.html
    """
    def __init__(self, **kwds):
        super().__init__(**kwds)
        if not 'MSK_DPAR_INTPNT_CO_TOL_DFEAS' in self.param:
            self.param['MSK_DPAR_INTPNT_CO_TOL_DFEAS'] = 1e-9

    @property
    def name(self):
        return "MOSEK"

    @property
    def type(self):
        return SolverType.SDR

    def solve(self, qcqp):  # pylint: disable=useless-super-delegation
        """
        Solve the given QCQP via its semidefinite relaxation using MOSEK.

        Parameters
        ----------
        qcqp : QCQP
            Specification of the quadratically-constrained quadratic problem.

        Returns
        -------
        result : QCQPResult
            Solution of the SDR associated with the QCQP.
        """
        # We solve the following problem:
        #
        #  min   c^T x   s.t.   Ax == b,  Bx <= d,  x_lb <= x <= x_ub
        #   x                   ...and a PSD constraint on the v_tld-part of x
        return super().solve(qcqp)

    @staticmethod
    def _add_voltage_matrix_constraints(task, qcqp, dim_x, ofs, x_lb, x_ub, x_bnd):
        """Add the PSD constraint on the ``v_tld``-part of ``x``."""
        # PSD constraint on V
        #
        # Consider the splitting of the Hermitian matrix V into its
        # real and imaginary part, i.e., V = V_r + j*V_i. Then,
        #
        #       V is PSD    <=>    V_cr = [ V_r  -V_i ] is PSD
        #                                 [ V_i   V_r ]
        #
        # This code implements the real-valued equivalent of V >= 0,
        # i.e., V_cr >= 0. To this end, a semidefinite matrix variable
        # V_cr is added and all relevant elements are equated to v_tld.
        # Note the following particularities:
        #
        #  1) MOSEK considers *symmetric* coefficient matrices and
        #     requires the specification of the *lower* triangular part.
        #  2) When equating the off-diagonal elements of V_cr to v_tld,
        #     the respective element of V_cr is taken twice (due to the
        #     sym. coeff. matrix), which explains the factor of 2 for
        #     the corresponding element in v_tld.
        #  3) V_i is skew-symmetric and, thus, its diagonal is zero.
        #     Correspondingly, the respective elements in V_cr are
        #     fixed at zero.
        #
        dim_v = qcqp.dim_v
        dim_V_cr = 2*dim_v
        K = qcqp.num_edges

        (e_src, e_dst) = qcqp.edges
        if np.any(e_src >= e_dst):
            raise RuntimeError("For simplicity, this solver assumes "
                               "that the source node index is smaller "
                               "than the destination node index in the "
                               "sparsity specification. At the time of "
                               "implementation, this was the default.")

        # Add the symmetric matrix V_cr
        task.appendbarvars([dim_V_cr])

        # Add constraints to equate v_tld to V_cr
        num_crt_V_cr = 3*dim_v + 4*K
        task.appendcons(num_crt_V_cr)
        task.putconboundlist(range(ofs, ofs + num_crt_V_cr),
                             [mosek.boundkey.fx] * num_crt_V_cr,
                             np.zeros(num_crt_V_cr),
                             np.zeros(num_crt_V_cr))

        # Diagonal of V_r in top-left block
        for i in range(dim_v):
            mtx_id = task.appendsparsesymmat(dim_V_cr, [i], [i], [-1])
            task.putaij(ofs + i, i, 1)
            task.putbaraij(ofs + i, 0, [mtx_id], [1.0])
        ofs += dim_v

        # Diagonal of V_r in bottom-right block
        for i in range(dim_v):
            mtx_id = task.appendsparsesymmat(dim_V_cr, [i + dim_v],
                                             [i + dim_v], [-1])
            task.putaij(ofs + i, i, 1)
            task.putbaraij(ofs + i, 0, [mtx_id], [1.0])
        ofs += dim_v

        # Diagonal of V_i in bottom-right block (fixed at zero)
        for i in range(dim_v):
            mtx_id = task.appendsparsesymmat(dim_V_cr,
                                             [i + dim_v], [i], [1])
            task.putbaraij(ofs + i, 0, [mtx_id], [1.0])
        ofs += dim_v

        # Off-diagonal elements of V_r in top-left block
        for k in range(K):
            mtx_id = task.appendsparsesymmat(dim_V_cr, [e_dst[k]],
                                             [e_src[k]], [-1])
            task.putaij(ofs + k, k + dim_v, 2)
            task.putbaraij(ofs + k, 0, [mtx_id], [1.0])
        ofs += K

        # Off-diagonal elements of V_r in bottom-right block
        for k in range(K):
            mtx_id = task.appendsparsesymmat(dim_V_cr, [e_dst[k] + dim_v],
                                             [e_src[k] + dim_v], [-1])
            task.putaij(ofs + k, k + dim_v, 2)
            task.putbaraij(ofs + k, 0, [mtx_id], [1.0])
        ofs += K

        # Upper-triangular part of V_i in bottom-left block
        for k in range(K):
            mtx_id = task.appendsparsesymmat(dim_V_cr, [e_src[k] + dim_v],
                                             [e_dst[k]], [-1])
            task.putaij(ofs + k, k + dim_v + K, 2)
            task.putbaraij(ofs + k, 0, [mtx_id], [1.0])
        ofs += K

        # Lower-triangular part of V_i in bottom-left block
        for k in range(K):
            mtx_id = task.appendsparsesymmat(dim_V_cr, [e_dst[k] + dim_v],
                                             [e_src[k]], [1])
            task.putaij(ofs + k, k + dim_v + K, 2)
            task.putbaraij(ofs + k, 0, [mtx_id], [1.0])
        ofs += K



#pylint: disable=unused-argument,anomalous-backslash-in-string
"""
IPOPT-based solver for the *hynet*-specific QCQP problem.
"""

import logging

import numpy as np
import scipy.sparse
import ipopt

from hynet.types_ import (hynet_int_,
                          hynet_float_,
                          hynet_complex_,
                          SolverType,
                          SolverStatus)
from hynet.utilities.worker import workers
from hynet.utilities.base import partition_iterable, Timer
from hynet.qcqp.solver import SolverInterface
from hynet.qcqp.problem import QCQPPoint
from hynet.qcqp.result import QCQPResult

_log = logging.getLogger(__name__)


def parse_status(status):
    """
    Parse the IPOPT status into a ``SolverStatus``.

    *Remark:* The interpretation of the status values can be deduced from
    the CyIPOPT code in [1]_.

    References
    ----------
    .. [1] https://github.com/matthias-k/cyipopt/blob/master/src/cyipopt.pyx
    """
    if status == 0:
        return SolverStatus.SOLVED
    if status == 1:
        return SolverStatus.INACCURATE
    if status == 2:
        return SolverStatus.INFEASIBLE
    return SolverStatus.FAILED


class QCQPSolver(SolverInterface):
    """
    IPOPT-based solver for the QCQP.

    This interface is based on CyIPOPT [1]_, a Python wrapper for IPOPT [2]_.

    For information on solver-specific parameters, see e.g. [3]_.

    See Also
    --------
    hynet.qcqp.solver.SolverInterface : Abstract base class for QCQP solvers.

    References
    ----------
    .. [1] https://github.com/matthias-k/cyipopt
    .. [2] https://projects.coin-or.org/Ipopt
    .. [3] https://www.coin-or.org/Ipopt/documentation/node40.html
    """

    def __init__(self, **kwds):
        super().__init__(**kwds)
        # REMARK: Experimentally, we found that this default setting provides a
        # good compromise between accuracy and performance.
        if 'tol' not in self.param:
            self.param['tol'] = 1e-7

    @property
    def name(self):
        return "IPOPT"

    @property
    def type(self):
        return SolverType.QCQP

    def solve(self, qcqp):
        """
        Solve the given QCQP using IPOPT.

        Parameters
        ----------
        qcqp : QCQP
            Specification of the quadratically-constrained quadratic problem.

        Returns
        -------
        result : QCQPResult
            Solution of the QCQP (in general only a *locally* optimal solution).
        """
        timer = Timer()

        problem = IPOPTProblem(qcqp, options=self)
        _log.debug("QCQP ~ Modeling ({:.3f} sec.)".format(timer.interval()))

        status, optimizer, dv_lb, dv_ub, dv_eq, dv_ineq, optimal_value = problem.solve()
        solver_time = timer.interval()
        _log.debug("QCQP ~ Solver ({:.3f} sec.)".format(solver_time))

        if status in [SolverStatus.SOLVED, SolverStatus.INACCURATE]:
            optimizer.v[:] = self.adjust_absolute_phase(qcqp, optimizer.v)
            result = QCQPResult(qcqp, self, status, solver_time,
                                optimizer=optimizer,
                                V=None,
                                optimal_value=optimal_value,
                                dv_lb=dv_lb,
                                dv_ub=dv_ub,
                                dv_eq=dv_eq,
                                dv_ineq=dv_ineq)
        else:
            result = QCQPResult(qcqp, self, status, solver_time)

        _log.debug("QCQP ~ Total time for solver ({:.3f} sec.)"
                   .format(timer.total()))
        return result


class IPOPTProblem(ipopt.problem):
    """
    Specification and solution of the QCQP via IPOPT.
    """
    # The following documents were helpful to understand the interface of (Cy)IPOPT:
    #   https://www.coin-or.org/Ipopt/documentation/node22.html
    #   https://github.com/matthias-k/cyipopt/blob/master/test/examplehs071.py
    #   https://github.com/matthias-k/cyipopt/blob/master/test/lasso.py
    #   https://pythonhosted.org/ipopt/tutorial.html
    #   https://pythonhosted.org/ipopt/reference.html

    def __init__(self, qcqp, options):
        """
        Initialize the problem formulation and prepare for the solution process.
        """
        (e_src, e_dst) = qcqp.edges
        if np.any(e_src >= e_dst):
            raise RuntimeError("For simplicity, this solver assumes "
                               "that the source node index is smaller "
                               "than the destination node index in the "
                               "sparsity specification. At the time of "
                               "implementation, this was the default.")

        self.qcqp = qcqp
        UNBOUNDED = 2.0e19  # Bounds are ignored if larger/smaller than +/- 10^19

        # Partitioning: x = [Re(v)^T, Im(v)^T, f^T, s^T, z^T]^T
        self._ofs_v_i = qcqp.dim_v
        self._ofs_f = self._ofs_v_i + qcqp.dim_v
        self._ofs_s = self._ofs_f + qcqp.dim_f
        self._ofs_z = self._ofs_s + qcqp.dim_s
        self._dim_x = self._ofs_z + qcqp.dim_z

        # Variable bounds
        x_lb = np.concatenate((-qcqp.ub.v,
                               -qcqp.ub.v,
                               qcqp.lb.f,
                               qcqp.lb.s,
                               qcqp.lb.z))
        x_ub = np.concatenate((qcqp.ub.v,
                               qcqp.ub.v,
                               qcqp.ub.f,
                               qcqp.ub.s,
                               qcqp.ub.z))
        self._idx_lb_nan = np.isnan(x_lb)
        self._idx_ub_nan = np.isnan(x_ub)
        x_lb[self._idx_lb_nan] = -UNBOUNDED
        x_ub[self._idx_ub_nan] = UNBOUNDED

        # Set phase to zero at root nodes by fixing the imaginary part at zero
        # (This avoids solution ambiguity w.r.t. an absolute phase shift in
        #  ``v`` for those components with a given root node and, therewith,
        #  can improve the solver convergence.)
        if qcqp.roots.size:
            idx_v_i_root = qcqp.roots + qcqp.dim_v
            x_lb[idx_v_i_root] = 0
            x_ub[idx_v_i_root] = 0

        # Initial point
        if qcqp.initial_point is not None:
            self._x0 = np.concatenate((qcqp.initial_point.v.real,
                                       qcqp.initial_point.v.imag,
                                       qcqp.initial_point.f,
                                       qcqp.initial_point.s,
                                       qcqp.initial_point.z))
        else:
            self._x0 = np.zeros(self._dim_x, dtype=hynet_float_)
            self._x0[:self._ofs_v_i] = (qcqp.lb.v + qcqp.ub.v) / 2

        # Objective: Part of gradient related to f, s, and z
        self._grad_f_fsz = scipy.sparse.vstack((qcqp.obj_func.c,
                                                qcqp.obj_func.a,
                                                qcqp.obj_func.r)
                                               ).toarray().reshape(1, -1)[0]

        # Objective: Hessian
        self._hessian_idx = self._get_hessian_index(qcqp)
        self._hessian_obj = self._get_hessian_data(qcqp, qcqp.obj_func)

        # Constraints and their upper and lower bounds:
        #  (1) Vectorize the constraints for efficient evaluation in the callback
        #  (2) Stack the equality and inequality constraints
        A, b, B, d = qcqp.get_constraint_vectorization()
        self._A = scipy.sparse.vstack((A, B)).tocsr()  # CSR for slicing & mat. prod.
        crt_b = np.concatenate((b, d))
        constraints = np.concatenate((qcqp.eq_crt, qcqp.ineq_crt))
        crt_ub = crt_b
        crt_lb = np.array(crt_b)
        crt_lb[len(qcqp.eq_crt):] = -UNBOUNDED

        # Column offsets in A with respect to the variables (vectorization!)
        ofs_A_v_real = qcqp.dim_v
        ofs_A_v_imag = ofs_A_v_real + qcqp.num_edges
        ofs_A_f = ofs_A_v_imag + qcqp.num_edges

        # Constraints: Jacobian
        #  (1) Create the stacking of c, a, and r for all constraints
        #  (2) Prepare the computation of the part of the gradient related to v
        #  (3) Stack the index data from (1) and (2) to defined the sparsity
        #      pattern of the Jacobian
        jac_fsz_mtx = self._A[:, ofs_A_f:].tocoo()
        jac_fsz_idx = (jac_fsz_mtx.row, jac_fsz_mtx.col + self._ofs_f)
        self._jacobian_fsz = jac_fsz_mtx.data
        jac_v_idx, self._jacobian_C = \
            self._get_jacobian_v(constraints, qcqp.dim_v)
        self._jacobian_idx = (np.concatenate((jac_v_idx[0], jac_fsz_idx[0])),
                              np.concatenate((jac_v_idx[1], jac_fsz_idx[1])))

        # Constraints: Hessian
        #   The following code performs the same as _get_hessian_data (cf.
        #   its application to the objective function above), as if it were
        #   applied to all constraints and the resulting Hessian data vectors
        #   were stacked horizontally. However, for performance reasons this
        #   implementation utilizes the precomputed stacking of the constraint
        #   vectorization above. Note that during the vectorization the
        #   off-diagonal elements are already multiplied by 2, thus only the
        #   diagonal elements need to be scaled.
        C_diag = 2*self._A[:, :ofs_A_v_real]
        C_offd_real = self._A[:, ofs_A_v_real:ofs_A_v_imag]
        C_offd_imag = self._A[:, ofs_A_v_imag:ofs_A_f]
        self._hessian_crt = scipy.sparse.hstack((C_diag,
                                                 C_diag,
                                                 C_offd_real,
                                                 C_offd_real,
                                                 C_offd_imag,
                                                 -C_offd_imag)).transpose().tocsr()

        # Initialize base class
        # (REMARK: This calls ``jacobianstructure`` and ``hessianstructure``)
        super().__init__(n=self._dim_x,
                         m=len(constraints),
                         problem_obj=self,
                         lb=x_lb,
                         ub=x_ub,
                         cl=crt_lb,
                         cu=crt_ub)

        # Set options
        self.addOption(b'sb', b'yes')  # Suppress the IPOPT banner
        self.addOption(b'print_level', 5 if options.verbose else 0)
        # self.addOption(b'print_options_documentation', b'yes')
        for key, value in options.param.items():
            self.addOption(key.encode(encoding='utf-8'),
                           value.encode(encoding='utf-8')
                           if isinstance(value, str) else value)

    def _split(self, x):
        """
        Split ``x = [Re(v)^T, Im(v)^T, f^T, s^T, z^T]^T`` into ``p = (v, f, s, z)``.

        Parameters
        ----------
        x : numpy.ndarray
            Optimization variable.

        Returns
        -------
        p : QCQPPoint
            Corresponding QCQP point ``p = (v, f, s, z)``.
        """
        return QCQPPoint(v=x[:self._ofs_v_i] + 1j*x[self._ofs_v_i:self._ofs_f],
                         f=x[self._ofs_f:self._ofs_s],
                         s=x[self._ofs_s:self._ofs_z],
                         z=x[self._ofs_z:])

    @staticmethod
    def _get_jacobian_v(constraints, dim_v):
        """
        Return the Jacobian sparsity pattern and evaluation support w.r.t. v.

        The gradient of the function
        ::
            v_cr^T C_cr v_cr + c^T f + a^T s + r^T z

        reads
        ::
            [ (2*C_cr v_cr)^T, c^T, a^T, r^T ]^T

        which considers the vectorization ``x = [v_cr^T, f^T, s^T, z^T]^T``
        and the composite real representation
        ::
            v_cr = [ Re(v)^T  Im(v)^T ]^T

        and
        ::
            C_cr = [ Re(C)  -Im(C) ]
                   [ Im(C)   Re(C) ]

        For the following discussion, it shall also be noted that
        ::
            (2*C_cr v_cr)^T = [ 2*Re(Cv)^T, 2*Im(Cv)^T ]

        The Jacobian of the vector-valued constraint function as queried by
        IPOPT is the vertical stacking of the transpose of the individual
        constraint function gradients. Let the constraint ``i`` have the
        parameters ``C_i``, ``c_i``, ``a_i``, and ``r_i``. Then, the Jacobian
        comprises the following blocks::

            [ 2*Re(C_1 v)^T  |  2*Im(C_1 v)^T  |  c_1^T  a_1^T  r_1^T ]
            [      ...       |       ...       |         ...          ]
            [ 2*Re(C_N v)^T  |  2*Im(C_N v)^T  |  c_N^T  a_N^T  r_N^T ]

            \_______________/ \_______________/ \____________________/
                   (1)               (2)                 (3)

        Block (3) remains constant for all iterations and is precomputed during
        initialization, while this method considers block (1) and (2). In row
        ``i``, the sparsity pattern in (1) and (2) is defined by the *nonzero
        rows* of ``C_i``. This method determines this sparsity pattern for all
        constraints and, alongside, returns a stacking of the row-reduced
        (i.e., only nonzero rows) constraint matrices ``C_i`` for an efficient
        evaluation of the nonzero elements in (1) and (2) during the callback.
        """
        timer = Timer()

        result_list = workers.map(_get_sparsity_and_row_reduction, constraints)

        list_row = []
        list_col = []
        list_jacobian_C = []
        for n, result in enumerate(result_list):
            if result is None:
                continue
            col, jacobian_C = result
            list_row.append(n + np.zeros(len(col), dtype=hynet_int_))
            list_col.append(col)
            list_jacobian_C.append(jacobian_C)

        row = np.concatenate(list_row)
        row = np.concatenate((row, row))          # For grad. w.r.t. v_r and v_i

        col = np.concatenate(list_col)
        col = np.concatenate((col, col + dim_v))  # For grad. w.r.t. v_r and v_i

        # Stacking is slow: Use multiprocessing to prestack, if available...
        if workers.is_multiprocessing and \
                len(list_jacobian_C) >= workers.num_workers:
            list_jacobian_C = workers.map(scipy.sparse.vstack,
                                          partition_iterable(list_jacobian_C,
                                                             workers.num_workers)
                                          )
        jacobian_C = scipy.sparse.vstack(list_jacobian_C).tocsr()

        _log.debug("QCQP ~ Preparation of v-part in Jacobian ({:.3f} sec.)"
                   .format(timer.total()))
        return (row, col), jacobian_C

    @staticmethod
    def _get_hessian_index(qcqp):
        """
        Returns the row and column index tuple for the Hessian sparsity pattern.

        This method defines the sparsity pattern for ``_get_hessian_data``.

        **Remark:** IPOPT requires that *only the lower left triangular part*
        (or, alternatively, the upper triangular part) is specified.

        **Remark:** This method as well as ``_get_hessian_data`` assumes that
        ``e_src >= e_dst``, where ``(e_src, e_dst) = qcqp.edges``. This is
        ensured during initialization.
        """
        (e_src, e_dst) = qcqp.edges
        range_v = np.arange(qcqp.dim_v)

        row = np.concatenate((range_v,               # Diag. top-left
                              range_v + qcqp.dim_v,  # Diag. bottom-right
                              e_dst,                 # Off-diag. top-left
                              e_dst + qcqp.dim_v,    # Off-diag. bottom-right
                              e_src + qcqp.dim_v,    # Off-diag. bottom-left
                              e_dst + qcqp.dim_v))   # Off-diag. bottom-left

        col = np.concatenate((range_v,               # Diag. top-left
                              range_v + qcqp.dim_v,  # Diag. bottom-right
                              e_src,                 # Off-diag. top-left
                              e_src + qcqp.dim_v,    # Off-diag. bottom-right
                              e_dst,                 # Off-diag. bottom-left
                              e_src))                # Off-diag. bottom-left
        return row, col

    @staticmethod
    def _get_hessian_data(qcqp, func):
        """
        Returns the Hessian as a data vector for the given function.

        The Hessian of the function
        ::
            v_cr^T C_cr v_cr + c^T f + a^T s + r^T z

        comprises ``2*C_cr`` in the top-left block considering the vectorization
        ``x = [v_cr^T, f^T, s^T, z^T]^T`` and the composite real representation
        ::
            v_cr = [ Re(v)^T  Im(v)^T ]^T

        and
        ::
            C_cr = [ Re(C)  -Im(C) ]
                   [ Im(C)   Re(C) ]

        The data is returned w.r.t. the sparsity pattern for the Hessian matrix
        as defined by ``_get_hessian_index``, i.e., the diagonal followed by
        the elements of the top-left, bottom-right, and bottom-left block of
        ``C_cr``. The block off-diagonal elements consider the sparsity pattern
        of the constraint matrix.
        """
        list_data = []
        (e_src, e_dst) = qcqp.edges

        C_diag = func.C.diagonal().real
        if e_src.size:
            C_offd = np.asarray(func.C.tocsr()[e_src, e_dst])[0]
        else:
            C_offd = np.ndarray(0, dtype=hynet_complex_)
        C_offd_real = C_offd.real
        C_offd_imag = C_offd.imag

        # REMARK: The real part of C is symmetric, while the imaginary part
        # of C is skew-symmetric.

        # Diagonal of top-left and bottom-right block
        list_data.append(C_diag)
        list_data.append(C_diag)

        # Off-diagonal elements (lower triangular part) of top-left block
        list_data.append(C_offd_real)

        # Off-diagonal elements (lower triangular part) of bottom-right block
        list_data.append(C_offd_real)

        # Off-diagonal elements of bottom-left block
        # (REMARK: The diagonal of the bottom-left block is zero)
        list_data.append(C_offd_imag)
        list_data.append(-C_offd_imag)

        return 2*np.concatenate(list_data)

    def solve(self):
        """Execute the interior-point solver."""
        logger = logging.getLogger()  # Root logger
        logger_disabled = logger.disabled
        logger.disabled = True  # CyIPOPT spams the log, so disable temporarily
        x, info = super().solve(self._x0)
        logger.disabled = logger_disabled
        status = parse_status(info['status'])

        if status in [SolverStatus.SOLVED, SolverStatus.INACCURATE]:
            optimizer = self._split(x)

            dv_lb = info['mult_x_L']
            dv_lb[self._idx_lb_nan] = np.nan
            dv_lb = self._split(dv_lb)
            dv_lb.v = None

            dv_ub = info['mult_x_U']
            dv_ub[self._idx_ub_nan] = np.nan
            dv_ub = self._split(dv_ub)
            dv_ub.v = None

            dv = info['mult_g']
            dv_eq = dv[:len(self.qcqp.eq_crt)]
            dv_ineq = dv[len(self.qcqp.eq_crt):]
            optimal_value = info['obj_val']
        else:
            optimizer = dv_lb = dv_ub = dv_eq = dv_ineq = optimal_value = None

        return status, optimizer, dv_lb, dv_ub, dv_eq, dv_ineq, optimal_value

    def objective(self, x):
        """
        Callback for the calculation of the objective function.
        """
        return self.qcqp.obj_func.evaluate(self._split(x))

    def gradient(self, x):
        """
        Callback for the calculation of the gradient of the objective function.
        """
        p = self._split(x)
        grad_f_v = 2*self.qcqp.obj_func.C.dot(p.v)
        return np.concatenate((grad_f_v.real, grad_f_v.imag, self._grad_f_fsz))

    def constraints(self, x):
        """Callback for the calculation of the constraints."""
        (e_src, e_dst) = self.qcqp.edges
        p = self._split(x)
        V_diag = np.square(np.abs(p.v))
        V_offd = np.multiply(p.v[e_src], p.v[e_dst].conj())
        x_tld = np.concatenate((V_diag, V_offd.real, V_offd.imag, p.f, p.s, p.z))
        return self._A.dot(x_tld)

    def jacobianstructure(self):
        """Definition of the sparsity structure of the Jacobian."""
        return self._jacobian_idx

    def jacobian(self, x):
        """Callback for the calculation of the Jacobian."""
        p = self._split(x)
        sparse_grad_g_v = 2*self._jacobian_C.dot(p.v)
        return np.concatenate([sparse_grad_g_v.real,
                               sparse_grad_g_v.imag,
                               self._jacobian_fsz])

    def hessianstructure(self):
        """
        Definition of the sparsity structure of the Hessian.

        *Remark:* If ``hessianstructure`` and ``hessian`` is *not* overridden,
        the Hessian is approximated.
        """
        return self._hessian_idx

    def hessian(self, x, lagrange, obj_factor):
        """Callback for the calculation of the Hessian."""
        return obj_factor * self._hessian_obj + self._hessian_crt.dot(lagrange)

    def intermediate(self, alg_mod, iter_count, obj_value, inf_pr, inf_du, mu,
                     d_norm, regularization_size, alpha_du, alpha_pr, ls_trials):
        """Intermediate callback."""
        return True  # Keep iterating...


def _get_sparsity_and_row_reduction(crt):
    """
    Return the index of nonzero rows and the row reduction of ``crt.C``.

    See Also
    --------
    IPOPTProblem._get_jacobian_v
    """
    if crt.C is not None:
        nonzero_rows = np.unique(crt.C.row)
        return nonzero_rows, crt.C.tocsr()[nonzero_rows, :]
    else:
        return None



"""
PYOMO-based solver for the *hynet*-specific QCQP problem.
"""

import logging

import numpy as np
import pyomo.environ as pe
from pyomo.opt import (SolverStatus as PyomoSolverStatus,
                       TerminationCondition as PyomoTerminationCondition)

from hynet.types_ import hynet_float_, SolverType, SolverStatus
from hynet.utilities.base import Timer
from hynet.qcqp.solver import SolverInterface
from hynet.qcqp.problem import QCQPPoint
from hynet.qcqp.result import QCQPResult

_log = logging.getLogger(__name__)


def parse_status(status, termination):
    """Parse the PYOMO status into a ``SolverStatus``."""
    # Accessing solver status and termination conditions:
    # http://www.pyomo.org/blog/2015/1/8/accessing-solver
    if status == PyomoSolverStatus.ok and \
            termination == PyomoTerminationCondition.optimal:
        return SolverStatus.SOLVED
    if termination == PyomoTerminationCondition.infeasible:
        return SolverStatus.INFEASIBLE
    if termination == PyomoTerminationCondition.unbounded:
        return SolverStatus.UNBOUNDED
    return SolverStatus.FAILED


def _add_lower_bounds(constraint_list, variable, lb):
    """Add variable lower bound constraints and return the resp. indices."""
    index = np.where(~np.isnan(lb))[0]
    for i in index:
        constraint_list.add(variable[i] >= lb[i])
    return index


def _add_upper_bounds(constraint_list, variable, ub):
    """Add variable upper bound constraints and return the resp. indices."""
    index = np.where(~np.isnan(ub))[0]
    for i in index:
        constraint_list.add(variable[i] <= ub[i])
    return index


def _func2pyomo(func, model):
    terms = []
    if func.C is not None:
        if func.C.row.size:
            (v_r, v_i) = (model.v_r, model.v_i)
            terms += [C_ij.real * (v_r[i]*v_r[j] + v_i[i]*v_i[j]) +
                      C_ij.imag * (v_i[i]*v_r[j] - v_r[i]*v_i[j])
                      for C_ij, i, j in zip(func.C.data, func.C.row, func.C.col)]
    if func.c is not None:
        terms += [c_i * model.f[i] for c_i, i in zip(func.c.data, func.c.row)]
    if func.a is not None:
        terms += [a_i * model.s[i] for a_i, i in zip(func.a.data, func.a.row)]
    if func.r is not None:
        terms += [r_i * model.z[i] for r_i, i in zip(func.r.data, func.r.row)]
    return sum(terms)


def _get_result(variable, index):
    return np.array([variable[i].value for i in index], dtype=hynet_float_)


def _get_duals(model, constraint_list, index):
    return np.array([model.dual[constraint_list[i]] for i in index],
                    dtype=hynet_float_)


class QCQPSolver(SolverInterface):
    """
    PYOMO-based solver for the QCQP.

    Parameters
    ----------
    solver_name : str, optional
        Name of the solver, see pyomo.environ.SolverFactory. This solver
        interface was only tested with ``ipopt``, which is set by default.

    See Also
    --------
    hynet.qcqp.solver.SolverInterface : Abstract base class for QCQP solvers.
    """

    def __init__(self, solver_name='ipopt', **kwds):
        super().__init__(**kwds)
        self.solver_name = solver_name

    @property
    def name(self):
        return "PYOMO + " + self.solver_name.upper()

    @property
    def type(self):
        return SolverType.QCQP

    def solve(self, qcqp):
        """
        Solve the given QCQP using PYOMO.

        Parameters
        ----------
        qcqp : QCQP
            Specification of the quadratically-constrained quadratic problem.

        Returns
        -------
        result : QCQPResult
            Solution of the QCQP (typically only a *locally* optimal solution).
        """
        timer = Timer()

        model = pe.ConcreteModel()
        model.dual = pe.Suffix(direction=pe.Suffix.IMPORT)  # For dual variables

        model.range_v = range(qcqp.dim_v)
        model.range_f = range(qcqp.dim_f)
        model.range_s = range(qcqp.dim_s)
        model.range_z = range(qcqp.dim_z)

        if qcqp.initial_point is not None:
            v0 = qcqp.initial_point.v
            f0 = qcqp.initial_point.f
            s0 = qcqp.initial_point.s
            z0 = qcqp.initial_point.z
        else:
            v0 = (qcqp.lb.v + qcqp.ub.v) / 2
            f0 = np.zeros(qcqp.dim_f, dtype=hynet_float_)
            s0 = np.zeros(qcqp.dim_s, dtype=hynet_float_)
            z0 = np.zeros(qcqp.dim_z, dtype=hynet_float_)

        model.v_r = pe.Var(model.range_v,
                           domain=pe.Reals,
                           dense=True,
                           initialize=lambda model, i: v0[i].real)
        model.v_i = pe.Var(model.range_v,
                           domain=pe.Reals,
                           dense=True,
                           initialize=lambda model, i: v0[i].imag)
        model.f = pe.Var(model.range_f,
                         domain=pe.Reals,
                         dense=True,
                         initialize=lambda model, i: f0[i])
        model.s = pe.Var(model.range_s,
                         domain=pe.Reals,
                         dense=True,
                         initialize=lambda model, i: s0[i])
        model.z = pe.Var(model.range_z,
                         domain=pe.Reals,
                         dense=True,
                         initialize=lambda model, i: z0[i])

        model.f_lb = pe.ConstraintList()
        model.s_lb = pe.ConstraintList()
        model.z_lb = pe.ConstraintList()
        model.idx_lb = QCQPPoint(
            v=None,
            f=_add_lower_bounds(model.f_lb, model.f, qcqp.lb.f),
            s=_add_lower_bounds(model.s_lb, model.s, qcqp.lb.s),
            z=_add_lower_bounds(model.z_lb, model.z, qcqp.lb.z)
        )

        model.f_ub = pe.ConstraintList()
        model.s_ub = pe.ConstraintList()
        model.z_ub = pe.ConstraintList()
        model.idx_ub = QCQPPoint(
            v=None,
            f=_add_upper_bounds(model.f_ub, model.f, qcqp.ub.f),
            s=_add_upper_bounds(model.s_ub, model.s, qcqp.ub.s),
            z=_add_upper_bounds(model.z_ub, model.z, qcqp.ub.z)
        )

        # Set phase to zero at root nodes by fixing the imaginary part at zero
        # (This avoids solution ambiguity w.r.t. an absolute phase shift in
        #  ``v`` for those components with a given root node and, therewith,
        #  can improve the solver convergence.)
        if qcqp.roots.size:
            model.root_phase = pe.ConstraintList()
            for root in qcqp.roots:
                model.root_phase.add(model.v_i[root] == 0)

        model.eq_crt = pe.ConstraintList()
        for crt in qcqp.eq_crt:
            model.eq_crt.add(_func2pyomo(crt, model) == crt.b)

        model.ineq_crt = pe.ConstraintList()
        for crt in qcqp.ineq_crt:
            model.ineq_crt.add(_func2pyomo(crt, model) <= crt.b)

        model.objective = pe.Objective(expr=_func2pyomo(qcqp.obj_func, model),
                                       sense=pe.minimize)

        solver = pe.SolverFactory(self.solver_name, tee=self.verbose)

        # Set solver parameters
        for key, value in self.param.items():
            solver.options[key] = value

        _log.debug("QCQP ~ Modeling ({:.3f} sec.)".format(timer.interval()))
        results = solver.solve(model, tee=self.verbose)
        solver_time = timer.interval()
        _log.debug("QCQP ~ Solver ({:.3f} sec.)".format(solver_time))

        status = parse_status(results.solver.status,
                              results.solver.termination_condition)

        if status in [SolverStatus.SOLVED, SolverStatus.INACCURATE]:
            v = _get_result(model.v_r, model.v_r_index) \
                + 1j * _get_result(model.v_i, model.v_i_index)
            optimizer = QCQPPoint(v=self.adjust_absolute_phase(qcqp, v),
                                  f=_get_result(model.f, model.range_f),
                                  s=_get_result(model.s, model.range_s),
                                  z=_get_result(model.z, model.range_z))
            optimal_value = model.objective()

            dv_eq = _get_duals(model, model.eq_crt, model.eq_crt_index)
            dv_ineq = _get_duals(model, model.ineq_crt, model.ineq_crt_index)

            # REMARK: Strangely, in some configurations PYOMO returns the dual
            # variables with their sign flipped. For example, on a Ubuntu
            # machine with PYOMO 5.5.0 and ipopt_bin 3.7.1 (At this date the
            # latest version for Linux) the dual variables were positive, but
            # on MAC OS X with PYOMO 5.5.0 and ipopt_bin 3.11.1 they were
            # negated. As the sign of dual variables of inequality constraints
            # is fixed, we use this to detect this strange behavior.
            if np.all(dv_ineq == 0):
                _log.warning("Dual variable sign detection failed as all "
                             "inequality dual variables are zero.")
            elif np.all(dv_ineq <= 0):
                dv_eq *= -1
                dv_ineq *= -1

            def create_bound_dual(index, dv_f, dv_s, dv_z):
                point = QCQPPoint(
                    v=None,
                    f=np.nan * np.ones(qcqp.dim_f, dtype=hynet_float_),
                    s=np.nan * np.ones(qcqp.dim_s, dtype=hynet_float_),
                    z=np.nan * np.ones(qcqp.dim_z, dtype=hynet_float_))
                point.f[index.f] = np.abs(dv_f)  # See remark above regarding
                point.s[index.s] = np.abs(dv_s)  # the use of the absolute
                point.z[index.z] = np.abs(dv_z)  # value.
                return point

            dv_lb = create_bound_dual(
                index=model.idx_lb,
                dv_f=_get_duals(model, model.f_lb, model.f_lb_index),
                dv_s=_get_duals(model, model.s_lb, model.s_lb_index),
                dv_z=_get_duals(model, model.z_lb, model.z_lb_index)
            )

            dv_ub = create_bound_dual(
                index=model.idx_ub,
                dv_f=_get_duals(model, model.f_ub, model.f_ub_index),
                dv_s=_get_duals(model, model.s_ub, model.s_ub_index),
                dv_z=_get_duals(model, model.z_ub, model.z_ub_index)
            )

            result = QCQPResult(qcqp, self, status, solver_time,
                                optimizer=optimizer,
                                V=None,
                                optimal_value=optimal_value,
                                dv_lb=dv_lb,
                                dv_ub=dv_ub,
                                dv_eq=dv_eq,
                                dv_ineq=dv_ineq)
        else:
            result = QCQPResult(qcqp, self, status, solver_time)

        _log.debug("QCQP ~ Result creation ({:.3f} sec.)"
                   .format(timer.interval()))
        _log.debug("QCQP ~ Total time for solver ({:.3f} sec.)"
                   .format(timer.total()))
        return result



"""
Solvers for the *hynet*-specific QCQP problem.
"""

import logging
_log = logging.getLogger(__name__)

# Build a list comprising all available solvers. Import in the order of
# preference, as the first element is used as the default solver.
AVAILABLE_SOLVERS = []

try:
    from hynet.solver import ipopt

    AVAILABLE_SOLVERS.append(ipopt.QCQPSolver)
except ImportError:
    pass

try:
    from hynet.solver import mosek
    from hynet.utilities.base import validate_mosek_license

    if validate_mosek_license():
        AVAILABLE_SOLVERS.append(mosek.SOCRSolver)
        AVAILABLE_SOLVERS.append(mosek.SDRSolver)
except ImportError:
    pass

try:
    from hynet.solver import picos

    AVAILABLE_SOLVERS.append(picos.SOCRSolver)
    AVAILABLE_SOLVERS.append(picos.SDRSolver)
except ImportError:
    pass

try:
    from hynet.solver import pyomo

    AVAILABLE_SOLVERS.append(pyomo.QCQPSolver)
except ImportError:
    pass

if not AVAILABLE_SOLVERS:
    _log.warning("No supported solvers were found.")



"""
Representation of a capability region of injectors and converters.
"""

from collections import namedtuple

import numpy as np

from hynet.types_ import hynet_float_


class HalfSpace(namedtuple('HalfSpace', ['offset', 'slope'])):
    """
    Representation of a half-space in a capability region.

    The half-spaces of a capability region are specified in terms of a
    nonnegative relative offset "``offset``" w.r.t. to the corresponding
    reactive power limit (i.e., the absolute offset is the reactive power limit
    multiplied by ``offset``) and its slope "``slope``".  In the dimension of
    active power, it is anchored at the corresponding active power limit.
    """
    def __repr__(self):
        return (u'({0:.0%}/{1:.0f}\N{DEGREE SIGN})'
                ).format(self.offset, np.arctan(self.slope) * 180/np.pi)


class CapRegion:
    """
    Specification of a capability region.

    The capability region is the intersection of a box and a polyhedron. Let
    ``s = [p, q]^T`` be the vector of active and reactive power injection. The
    capability region ``S`` is the intersection of ``S_box`` and ``S_phs``,
    where
    ::

        (1) S_box = { s in R^2: p_min <= p <= p_max, q_min <= q <= q_max }
        (2) S_phd = { s in R^2: A*s <= b }.

    The polyhedron ``S_phd`` is defined by the properties ``lt``, ``lb``,
    ``rt``, and ``rb``, which are half-spaces that are defined by a
    *nonnegative* relative offset and a slope, see the parameter description
    below. The parametrization in ``(2)`` can be obtained using the method
    ``get_polyhedron``.

    Parameters
    ----------
    p_min : .hynet_float_
        Active power lower bound in MW.
    p_max : .hynet_float_
        Active power upper bound in MW.
    q_min : .hynet_float_
        Reactive power lower bound in Mvar.
    q_max : .hynet_float_
        Reactive power upper bound in Mvar.
    lt : HalfSpace or None
        Left-top half-space or ``None`` if omitted. This half-space is
        anchored at ``(p,q)`` with ``p = p_min`` and ``q = lt.offset * q_max``.
        The slope must be *positive*.
    rt : HalfSpace or None
        Right-top half-space or ``None`` if omitted. This half-space is
        anchored at ``(p,q)`` with ``p = p_max`` and ``q = rt.offset * q_max``.
        The slope must be *negative*.
    lb : HalfSpace or None
        Left-bottom half-space or ``None`` if omitted. This half-space is
        anchored at ``(p,q)`` with ``p = p_min`` and ``q = lb.offset * q_min``.
        The slope must be *negative*.
    rb : HalfSpace or None
        Right-bottom half-space or ``None`` if omitted. This half-space is
        anchored at ``(p,q)`` with ``p = p_max`` and ``q = rb.offset * q_min``.
        The slope must be *positive*.
    """
    def __init__(self, p_bnd=None, q_bnd=None,
                 lt=None, rt=None, lb=None, rb=None):
        (self.p_min, self.p_max) = (0, 0) if p_bnd is None else p_bnd
        (self.q_min, self.q_max) = (0, 0) if q_bnd is None else q_bnd
        self._lt = self._rt = self._lb = self._rb = None
        (self.lt, self.rt, self.lb, self.rb) = (lt, rt, lb, rb)

    @property
    def lt(self):
        """Return the left-top half-space."""
        return self._lt

    @lt.setter
    def lt(self, value):
        """Set the left-top half-space."""
        if isinstance(value, HalfSpace):
            if value.offset < 0:
                raise ValueError("Offset must be nonnegative.")
            elif value.slope <= 0:
                raise ValueError("Slope must be positive.")
        elif value is not None:
            raise ValueError("Instance of HalfSpace or None expected.")
        self._lt = value

    @property
    def rt(self):
        """Return the right-top half-space."""
        return self._rt

    @rt.setter
    def rt(self, value):
        """Set the right-top half-space"""
        if isinstance(value, HalfSpace):
            if value.offset < 0:
                raise ValueError("Offset must be nonnegative.")
            elif value.slope >= 0:
                raise ValueError("Slope must be negative.")
        elif value is not None:
            raise ValueError("Instance of HalfSpace or None expected.")
        self._rt = value

    @property
    def lb(self):
        """Return the left-bottom half-space."""
        return self._lb

    @lb.setter
    def lb(self, value):
        """Set the left-bottom half-space."""
        if isinstance(value, HalfSpace):
            if value.offset < 0:
                raise ValueError("Offset must be nonnegative.")
            elif value.slope >= 0:
                raise ValueError("Slope must be negative.")
        elif value is not None:
            raise ValueError("Instance of HalfSpace or None expected.")
        self._lb = value

    @property
    def rb(self):
        """Return the right-bottom half-space."""
        return self._rb

    @rb.setter
    def rb(self, value):
        """Set the right-bottom half-space."""
        if isinstance(value, HalfSpace):
            if value.offset < 0:
                raise ValueError("Offset must be nonnegative.")
            elif value.slope <= 0:
                raise ValueError("Slope must be positive.")
        elif value is not None:
            raise ValueError("Instance of HalfSpace or None expected.")
        self._rb = value

    def __repr__(self):
        rep = '[{0.p_min},{0.p_max}]x[{0.q_min},{0.q_max}]'.format(self)
        if self.lt is not None:
            rep += 'LT' + str(self.lt)
        if self.rt is not None:
            rep += 'RT' + str(self.rt)
        if self.lb is not None:
            rep += 'LB' + str(self.lb)
        if self.rb is not None:
            rep += 'RB' + str(self.rb)
        return rep

    def __eq__(self, other):
        """Return True if the capability regions feature the same parameters."""
        if other is None:
            return False
        return (self.p_min == other.p_min and
                self.p_max == other.p_max and
                self.q_min == other.q_min and
                self.q_max == other.q_max and
                self.lt == other.lt and
                self.rt == other.rt and
                self.lb == other.lb and
                self.rb == other.rb)

    def has_polyhedron(self):
        """Return ``True`` if any of the half-spaces is set."""
        return any(x is not None for x in [self.lt, self.rt, self.lb, self.rb])

    def get_polyhedron(self):
        """
        Returns the polyhedron formulation ``S_phd = { s in R^2: A*s <= b }``.

        Returns
        -------
        A : np.ndarray
            ``A`` in the formulation of the polyhedron above.
        b : np.ndarray
            ``b`` in the formulation of the polyhedron above.
        name : list[str]
            Abbreviation of the corresponding half-space for each row.
        """
        A = np.ndarray((0, 2), dtype=hynet_float_)
        b = np.ndarray(0, dtype=hynet_float_)
        name = []
        if self.lt is not None:
            (r, s) = (self.lt.offset, self.lt.slope)
            A = np.vstack((A, [-s, 1]))
            b = np.concatenate((b, [r * self.q_max - s * self.p_min]))
            name.append('lt')
        if self.rt is not None:
            (r, s) = (self.rt.offset, self.rt.slope)
            A = np.vstack((A, [-s, 1]))
            b = np.concatenate((b, [r * self.q_max - s * self.p_max]))
            name.append('rt')
        if self.lb is not None:
            (r, s) = (self.lb.offset, self.lb.slope)
            A = np.vstack((A, [s, -1]))
            b = np.concatenate((b, [s * self.p_min - r * self.q_min]))
            name.append('lb')
        if self.rb is not None:
            (r, s) = (self.rb.offset, self.rb.slope)
            A = np.vstack((A, [s, -1]))
            b = np.concatenate((b, [s * self.p_max - r * self.q_min]))
            name.append('rb')
        return A, b, name

    def scale(self, scaling):
        """Scale the upper and lower bound on active and reactive power."""
        if scaling < 0:
            raise ValueError("Expecting a nonnegative scaling factor.")
        self.p_min *= scaling
        self.p_max *= scaling
        self.q_min *= scaling
        self.q_max *= scaling
        return self

    def add_power_factor_limit(self, power_factor):
        """
        Add a left-top and left-bottom half-space to limit the power factor.

        Parameters
        ----------
        power_factor : .hynet_float_
            Power factor limit.
        """
        if not 0 < power_factor < 1:
            raise ValueError("Power factor must be in (0,1).")
        if not 0 <= self.p_min < self.p_max and self.q_min < 0 < self.q_max:
            raise ValueError(("Capability region must be in the right half-"
                              "plane, with a nonempty interior that includes "
                              "zero reactive power for all p_min < p < p_max."))
        slope = np.sqrt(1/np.square(power_factor) - 1)
        self.lt = HalfSpace(slope * self.p_min / self.q_max, slope)
        self.lb = HalfSpace(-slope * self.p_min / self.q_min, -slope)

    def edit(self):
        """
        Edit this capability region in the capability region visualizer.

        **Caution:** Due to technical reasons, all open ``matplotlib`` figures
        are closed during this call.

        **Remark:** In case you are using MAC OS X, please be aware of `this
        issue <https://stackoverflow.com/questions/32019556/
        matplotlib-crashing-tkinter-application>`_ with ``matplotlib`` and
        ``tkinter``, which causes Python to crash if the capability region
        visualizer is opened. To avoid it, set ``matplotlib``'s backend to
        ``TkAgg`` *before* importing *hynet* using

        >>> import matplotlib
        >>> matplotlib.use('TkAgg')
        """
        from hynet.visual.capability.visualizer import Window
        Window.show(self, edit=True)

    def show(self, operating_point=None):
        """
        Show this capability region in the capability region visualizer.

        **Caution:** Due to technical reasons, all open ``matplotlib`` figures
        are closed during this call.

        **Remark:** In case you are using MAC OS X, please be aware of `this
        issue <https://stackoverflow.com/questions/32019556/
        matplotlib-crashing-tkinter-application>`_ with ``matplotlib`` and
        ``tkinter``, which causes Python to crash if the capability region
        visualizer is opened. To avoid it, set ``matplotlib``'s backend to
        ``TkAgg`` *before* importing *hynet* using

        >>> import matplotlib
        >>> matplotlib.use('TkAgg')

        Parameters
        ----------
        operating_point : .hynet_complex_, optional
            If provided, a marker is shown for this operating point in the
            P/Q-plane.
        """
        from hynet.visual.capability.visualizer import Window
        pq_point = (operating_point.real, operating_point.imag)
        Window.show(self, edit=False, operating_point=pq_point)


class ConverterCapRegion(CapRegion):
    """
    Specification of a converter capability region.

    The capability region is specified in terms of the apparent power flow
    at the respective terminal bus of the converter, see ``CapRegion`` for
    more information. With respect to the parameters returned by
    ``get_polyhedron`` and ``get_box_constraint``, this class considers the
    converter state variable ``f = [p_src, p_dst, q_src, q_dst]^T`` instead
    of the apparent power vector ``s = [p, q]^T`` considered in ``CapRegion``.

    See Also
    --------
    CapRegion
    """
    def get_polyhedron(self, terminal, loss_factor):
        """
        Returns the polyhedron formulation ``S_phd = { f in R^4: A*f <= b }``.

        **Remark:** In contrast to ``CapRegion``, the polyhedron does not only
        include the half-spaces, but also the upper and lower bound on active
        power.

        Parameters
        ----------
        terminal : str
            Terminal of the converter (``'src'`` or ``'dst'``) with which the
            capability region is associated.
        loss_factor : float
            Proportional loss factor for (a) the *backward* flow if the
            terminal is ``'src'`` and (b) the *forward* flow if the terminal
            is ``'dst'``.

        Returns
        -------
        A : np.ndarray
            ``A`` in the formulation of the polyhedron above.
        b : np.ndarray
            ``b`` in the formulation of the polyhedron above.
        name : list[str]
            Abbreviation of the corresponding half-space or active power limit
            for each row.

        See Also
        --------
        get_box_constraint
        """
        if not 0 <= loss_factor < 1:
            raise ValueError("Loss factor must be in [0,1).")

        if terminal == 'src':
            F = [[1, loss_factor - 1, 0, 0], [0, 0, 1, 0]]
        elif terminal == 'dst':
            F = [[loss_factor - 1, 1, 0, 0], [0, 0, 0, 1]]
        else:
            raise ValueError("Expecting 'src' or 'dst' for terminal.")

        # Reformulate polyhedron in terms of f and add active power LB and UB
        F = np.array(F, dtype=hynet_float_)
        A, b, name = super().get_polyhedron()
        A = np.vstack((A.dot(F), F[0, :], -F[0, :]))
        b = np.concatenate((b, [self.p_max, -self.p_min]))
        name += ['p_max', 'p_min']
        return A, b, name

    @staticmethod
    def get_box_constraint(cap_src, cap_dst):
        """
        Return ``(f_lb, f_ub)`` of the state constraint ``f_lb <= f <= f_ub``.

        A converter comprises two capability regions, one at the source terminal
        (``cap_src``) and one at the destination terminal (``cap_dst``). Thus,
        the box constraint on the converter state variable is a combination of
        both and returned by this function. Note that the active power
        constraints are included in ``get_polyhedron`` and the upper bounds on
        ``p_src`` and ``p_dst`` in ``f`` are merely additional loosened
        upper bounds (cf. the remark in ``get_voltage_magnitude_bounds`` of
        ``SystemModel``).
        """
        f_lb = np.zeros(4, dtype=hynet_float_)  # LB on p_src, p_dst is 0
        f_lb[2] = cap_src.q_min  # q_src
        f_lb[3] = cap_dst.q_min  # q_dst

        f_ub = np.zeros(4, dtype=hynet_float_)
        f_ub[0] = np.max([abs(cap_src.p_max), abs(cap_dst.p_min)]) * 1.1  # p_src
        f_ub[1] = np.max([abs(cap_dst.p_max), abs(cap_src.p_min)]) * 1.1  # p_dst
        f_ub[2] = cap_src.q_max  # q_src
        f_ub[3] = cap_dst.q_max  # q_dst

        return f_lb, f_ub

    def add_power_factor_limit(self, power_factor):
        raise AttributeError(("The power factor limit is not applicable "
                              "to converter capability regions."))



#pylint: disable=too-many-instance-attributes,too-many-arguments
"""
Representation of the result of a *hynet*-specific QCQP.
"""

import logging

import numpy as np
import pandas as pd

from hynet.qcqp.problem import QCQPPoint  # pylint: disable=unused-import
from hynet.utilities.base import Timer
from hynet.utilities.rank1approx import calc_rank1approx_mse

_log = logging.getLogger(__name__)


class QCQPResult:
    """
    Solution of a quadratically constrained quadratic program.

    **Caution:** The constructor adjusts the optimal objective value, the
    optimizer, and the dual variables according to the scaling of the problem,
    i.e., the properties of the result correspond to the *unscaled* QCQP.

    Parameters
    ----------
    qcqp : QCQP
        Problem specification.
    solver : SolverInterface
        Solver object by which the result was obtained.
    solver_status : SolverStatus
        Status reported by the solver after performing the optimization.
    solver_time : float
        Duration of the call to the solver in seconds.
    optimizer : QCQPPoint or None
        Optimal optimization variable ``(v^*, f^*, s^*, z^*)`` or ``None`` if
        the solver failed. In case of a relaxation, ``v^*`` is the rank-1
        approximation of ``V^*``.
    V : .hynet_sparse_ or None
        Optimal optimization variable ``V^*`` in case of a relaxation or
        ``None`` if the solver failed or a nonconvex-QCQP solver was employed.
    optimal_value : float
        Optimal objective value or ``numpy.nan`` if the solver failed.
    reconstruction_mse : float
        The mean squared error of the reconstructed ``v^*`` in case of a
        relaxation or ``numpy.nan`` if the solver failed or a nonconvex-QCQP
        solver was employed.
    dv_lb : QCQPPoint or None
        Optimal dual variables of the lower bounds on ``f``, ``s``, and ``z``
        or ``None`` if the solver failed. The attribute ``v`` is set ``None``
        as these bounds are optional, cf. the QCQP specification.
    dv_ub : QCQPPoint or None
        Optimal dual variables of the upper bounds on ``f``, ``s``, and ``z``
        or ``None`` if the solver failed. The attribute ``v`` is set ``None``
        as these bounds are optional, cf. the QCQP specification.
    dv_eq : numpy.ndarray or None
        Optimal dual variables of the equality constraints or ``None`` if the
        solver failed.
    dv_ineq : numpy.ndarray or None
        Optimal dual variables of the inequality constraints or ``None`` if the
        solver failed.
    """
    def __init__(self, qcqp, solver, solver_status, solver_time,
                 optimizer=None, V=None, optimal_value=None,
                 dv_lb=None, dv_ub=None, dv_eq=None, dv_ineq=None):
        timer = Timer()
        self.qcqp = qcqp
        self.solver = solver
        self.solver_status = solver_status
        self.solver_time = solver_time
        self._optimizer_normalized = optimizer
        self.V = V
        self.optimal_value = optimal_value
        self.dv_lb = dv_lb
        self.dv_ub = dv_ub
        self.dv_eq = dv_eq
        self.dv_ineq = dv_ineq

        # Compensate the scaling of the objective, variables, and constraints
        #
        # REMARK: If a constraint is scaled by s, then the associated dual
        # variable scales by 1/s (cf. the first order optimality conditions).
        # Therefore, to obtain the dual variable of the unscaled constraint,
        # the dual variable of the scaled constraint must be multiplied by s.
        obj_scaling = qcqp.obj_func.scaling

        if self.optimal_value is None:
            self.optimal_value = np.nan
        self.optimal_value /= obj_scaling

        if self._optimizer_normalized is not None:
            scaling = qcqp.normalization.copy().reciprocal()
            self.optimizer = self._optimizer_normalized.copy().scale(scaling)
        else:
            self.optimizer = None

        if self.V is not None:
            self.V = self.V / qcqp.normalization.v ** 2  # New matrix!

        bound_dual_scaling = qcqp.normalization.copy().scale(1 / obj_scaling)
        if self.dv_lb is not None:
            self.dv_lb = self.dv_lb.copy().scale(bound_dual_scaling)

        if self.dv_ub is not None:
            self.dv_ub = self.dv_ub.copy().scale(bound_dual_scaling)

        if self.dv_eq is not None:
            self.dv_eq = self.dv_eq / obj_scaling  # New array!
            for i in range(len(qcqp.eq_crt)):
                self.dv_eq[i] *= qcqp.eq_crt[i].scaling

        if self.dv_ineq is not None:
            self.dv_ineq = self.dv_ineq / obj_scaling  # New array!
            for i in range(len(qcqp.ineq_crt)):
                self.dv_ineq[i] *= qcqp.ineq_crt[i].scaling

        # Calculate the reconstruction error in case of a relaxation
        if self.V is not None:
            self.reconstruction_mse = calc_rank1approx_mse(self.V,
                                                           self.optimizer.v,
                                                           qcqp.edges)
        else:
            self.reconstruction_mse = np.nan

        _log.debug("QCQP result creation ({:.3f} sec.)"
                   .format(timer.total()))

    @property
    def empty(self):
        """Return ``True`` if the QCQP result does not contain an optimizer."""
        return self.optimizer is None

    def get_result_tables(self, tables=None,
                          dual_prefix='dv_', value_prefix='cv_'):
        """
        Return a dictionary of data frames with the dual and constraint result.

        According to the table and ID information of the constraint objects,
        a dictionary of data frames is assembled that contains the dual
        variables and the *equality* constraint function values with the right
        hand side subtracted.
        """
        tables = {} if tables is None else tables

        if self.empty:
            return tables

        for crt, dv in zip(self.qcqp.eq_crt, self.dv_eq):
            if crt.table is None:
                continue
            if crt.table not in tables:
                tables[crt.table] = pd.DataFrame()
            tables[crt.table].at[crt.id, dual_prefix + crt.name] = dv
            tables[crt.table].at[crt.id, value_prefix + crt.name] = \
                crt.evaluate(self._optimizer_normalized) / crt.scaling

        for crt, dv in zip(self.qcqp.ineq_crt, self.dv_ineq):
            if crt.table is None:
                continue
            if crt.table not in tables:
                tables[crt.table] = pd.DataFrame()
            tables[crt.table].at[crt.id, dual_prefix + crt.name] = dv

        return tables



"""
Verification of a steady-state scenario.
"""

import logging

import numpy as np
import pandas as pd

from hynet.types_ import hynet_eps, BusType, BranchType
from hynet.utilities.base import Timer
from hynet.utilities.graph import get_graph_components

_log = logging.getLogger(__name__)


def verify_scenario(scr):
    """
    Verify the integrity and validity of the scenario.

    This function is actually part of the ``Scenario`` class. Due to its extent,
    it was moved to a separate module in order to improve code readability.

    Raises
    ------
    ValueError
        In case any kind of integrity or validity violation is detected.
    """
    timer = Timer()
    e_src = scr.e_src
    e_dst = scr.e_dst

    z_bar = scr.branch['z_bar']
    y_src = scr.branch['y_src']
    y_dst = scr.branch['y_dst']

    rho_src = scr.branch['rho_src']
    rho_dst = scr.branch['rho_dst']
    rho = pd.Series(np.multiply(rho_src.values.conj(), rho_dst.values),
                    index=scr.branch.index)
    rho_angle = pd.Series(np.angle(rho.values) * 180/np.pi,
                          index=scr.branch.index)

    base_kv = scr.bus['base_kv']

    angle_min = scr.branch['angle_min']
    angle_max = scr.branch['angle_max']

    if scr.base_mva <= 0:
        raise ValueError("The MVA base is invalid (positive number expected).")

    if scr.loss_price < 0:
        raise ValueError("The loss price is invalid (nonnegative number expected).")

    # Topological checks

    if scr.num_buses < 1:
        raise ValueError("The grid does not comprise any buses.")

    if scr.num_injectors < 1:
        raise ValueError("There is no injector connected to the grid.")

    if not (np.all(np.isin(scr.branch['src'], scr.bus.index)) and
            np.all(np.isin(scr.branch['dst'], scr.bus.index))):
        raise ValueError("Some branches connect to non-existing buses.")

    if not (np.all(np.isin(scr.converter['src'], scr.bus.index)) and
            np.all(np.isin(scr.converter['dst'], scr.bus.index))):
        raise ValueError("Some converters connect to non-existing buses.")

    if not np.all(np.isin(scr.injector['bus'], scr.bus.index)):
        raise ValueError("Some injectors connect to non-existing buses.")

    if np.any(np.equal(scr.branch['src'], scr.branch['dst'])):
        raise ValueError("The grid contains branch-based self-loops.")

    if np.any(np.logical_and(np.not_equal(base_kv.iloc[e_src],
                                          base_kv.iloc[e_dst]),
                             scr.branch['type'] != BranchType.TRANSFORMER)):
        raise ValueError("Some non-transformer branches connect buses "
                         "with a different base voltage.")

    if np.any(np.equal(scr.converter['src'], scr.converter['dst'])):
        raise ValueError("The grid contains converter-based self-loops.")

    for subgrid in get_graph_components(np.arange(scr.num_buses),
                                        (e_src.values, e_dst.values)):
        subgrid = np.sort(subgrid)
        num_ref = np.count_nonzero(scr.bus['ref'].iloc[subgrid])
        if num_ref > 1:
            _log.warning("Ambiguous references detected. Bus IDs: "
                         + str(list(scr.bus.iloc[subgrid].query('ref').index)))
        bus_type = scr.bus['type'].iloc[subgrid]
        is_ac = np.all(bus_type == BusType.AC)
        is_dc = np.all(bus_type == BusType.DC)
        if not (is_ac or is_dc):
            raise ValueError("Inconsistent bus types in the subgrid comprising "
                             "the buses " + str(list(scr.bus.index[subgrid])))
        if is_ac and num_ref < 1:
            # We only require AC subgrids to have a reference bus
            raise ValueError("Reference is missing in the AC subgrid comprising "
                             "the buses " + str(list(scr.bus.index[subgrid])))

    if not scr.has_acyclic_dc_subgrids():
        raise ValueError("The grid contains meshed DC subgrids. hynet only "
                         "supports radial DC subgrids.")

    # Electrical checks

    if np.any(z_bar.values.real < 0):
        _log.warning("The series resistance of some branches is negative.")

    if np.any(z_bar.values.real == 0):
        _log.info("The series resistance of some branches is zero. This "
                  "can be detrimental to the exactness of the relaxation.")

    if np.any(z_bar.values.imag < 0):
        _log.info("The series reactance of some branches is capacitive. This "
                  "can be detrimental to the exactness of the relaxation.")

    if np.any(z_bar.abs() < 1e-6):
        _log.warning("Some branches are close to an electrical short. "
                     "This leads to a bad conditioning of the OPF problem "
                     "and may cause numerical and/or convergence issues.")

    if np.any(y_src.values.real < 0) or np.any(y_dst.values.real < 0):
        raise ValueError("The shunt conductance of some branches is negative.")

    if np.any(np.multiply(y_src.abs(), z_bar.abs()) > 1) or \
            np.any(np.multiply(y_dst.abs(), z_bar.abs()) > 1):
        _log.info("Some branches are not properly insulated. This can "
                  "be detrimental to the exactness of the relaxation.")

    if np.any(rho.abs() == 0):
        raise ValueError("Some branches exhibit a total voltage ratio of zero.")

    if np.any(rho_angle.abs() > 90):
        raise ValueError("Some branches exhibit a total phase ratio of "
                         "more than 90 degrees.")

    for parallel_branches in scr._get_parallel_branch_indices():
        idx = e_src.values[parallel_branches] == e_src.values[parallel_branches[0]]
        rho_diff = np.concatenate((rho.values[parallel_branches[idx]],
                                   rho.values[parallel_branches[~idx]].conj()))
        rho_diff -= rho.values[parallel_branches[0]]
        if np.any(np.abs(rho_diff) > hynet_eps):
            _log.info("In the group "
                      + str(list(scr.branch.index[parallel_branches]))
                      + " of parallel branches the total voltage "
                        "ratios do not agree. This can be detrimental "
                        "to the branch utilization as well as the "
                        "exactness of the relaxation.")

    if np.any(scr.converter[['loss_fwd', 'loss_bwd']] < 0) or \
            np.any(scr.converter[['loss_fwd', 'loss_bwd']] >= 100):
        raise ValueError("Some converter loss factors are not within [0,100).")

    if np.any(scr.converter[['loss_fwd', 'loss_bwd']] == 0):
        _log.warning("Some converters are lossless. Due to modeling with "
                     "a forward and backward flow, this leads to "
                     "an ambiguity in the optimal converter state, which "
                     "may be detrimental to solver convergence.")

    # Electrical checks: Restrictions for DC grids

    dc_bus = scr.bus.loc[scr.bus['type'] == BusType.DC]

    if np.any(dc_bus['y_tld'] != 0):
        raise ValueError("Some bus shunts in the DC subgrids are nonzero.")

    if np.any(dc_bus['load'].imag != 0):
        raise ValueError("Some loads in the DC subgrids demand reactive power.")

    dc_branch = scr.branch.loc[
        scr.bus.loc[scr.branch['src'], 'type'].values == BusType.DC]

    if np.any(dc_branch['rho_src'] != 1) or np.any(dc_branch['rho_dst'] != 1):
        raise ValueError("Some transformers of the DC branches are non-unity.")

    if np.any(dc_branch['y_src'] != 0) or np.any(dc_branch['y_dst'] != 0):
        raise ValueError("Some shunt admittances of the DC branches are nonzero.")

    if np.any(dc_branch['z_bar'].real <= 0):
        raise ValueError("Some DC branches are lossless or active.")

    if np.any(dc_branch['z_bar'].imag != 0):
        raise ValueError("Some series impedances of the DC branches "
                         "exhibit a nonzero reactance.")

    # if np.any(~np.isnan(dc_branch['angle_min'])) or \
    #         np.any(~np.isnan(dc_branch['angle_max'])):
    #     _log.warning("Some DC branches include angle difference limits.")

    # if np.any(~np.isnan(dc_branch['drop_min'])) or \
    #         np.any(~np.isnan(dc_branch['drop_max'])):
    #     _log.warning("Some DC branches include voltage drop limits.")

    dc_converter_src = scr.converter.loc[
        scr.bus.loc[scr.converter['src'], 'type'].values == BusType.DC]
    dc_converter_dst = scr.converter.loc[
        scr.bus.loc[scr.converter['dst'], 'type'].values == BusType.DC]
    for id_, cap in zip(np.concatenate((dc_converter_src.index,
                                        dc_converter_dst.index)),
                        np.concatenate((dc_converter_src['cap_src'].values,
                                        dc_converter_dst['cap_dst'].values))):
        if cap.q_max != 0 or cap.q_min != 0:
            raise ValueError(("Converter {0} offers reactive power on the "
                              "DC side.").format(id_))

    dc_injector = scr.injector.loc[
        scr.bus.loc[scr.injector['bus'], 'type'].values == BusType.DC]
    for id_, cap in zip(dc_injector.index, dc_injector['cap'].values):
        if cap.q_max != 0 or cap.q_min != 0:
            raise ValueError(("Injector {0} offers reactive power to a "
                              "DC subgrid.").format(id_))

    # Constraint checks

    if not (np.all(scr.bus['v_min'] > 0) and
            np.all(scr.bus['v_max'] >= scr.bus['v_min'])):
        raise ValueError("Some voltage limits are infeasible, zero, or missing.")

    if np.any(scr.branch['rating'] <= 0):
        raise ValueError("Some branch ratings are infeasible or zero.")

    if np.any(angle_min < -89) or np.any(angle_max > 89) or \
            np.any(angle_min >= angle_max):
        raise ValueError("Some angle difference limits are infeasible, "
                         "equal, or not within [-89, 89] degrees")

    if np.any(angle_min > -rho_angle) or np.any(angle_max < -rho_angle):
        _log.info("Some angle difference constraints do not enclose "
                  "the negated total phase shift. This can be "
                  "detrimental to the exactness of the relaxation.")

    if np.any(scr.branch['drop_min'] < -100) or \
            np.any(scr.branch['drop_max'] <= scr.branch['drop_min']):
        raise ValueError("Some voltage drop limits are infeasible, "
                         "equal, or not within [-100, +inf).")

    for n, cap in enumerate(np.concatenate((scr.converter['cap_src'].values,
                                            scr.converter['cap_dst'].values,
                                            scr.injector['cap'].values))):
        # Note that only the box constraint needs to be checked, proper
        # specification of the half-spaces is ensured in CapRegion
        if not (cap.p_max >= cap.p_min and cap.q_max >= cap.q_min):
            raise ValueError(("Some {0:s} capability regions are infeasible "
                              "or incompletely specified.").format(
                             'converter'
                             if n < 2*scr.num_converters else 'injector'))
        if cap.has_polyhedron():
            if cap.p_max == cap.p_min and cap.q_max == cap.q_min != 0:
                _log.warning("Singleton capability regions with nonzero "
                             "reactive power and polyhedral constraints are "
                             "present. This potentially causes infeasibility.")
            elif not cap.q_min <= 0 <= cap.q_max:
                _log.info("Capability regions with polyhedral constraints "
                          "are present, where the reactive power limits "
                          "do not include zero. This is not recommended, "
                          "as the specification becomes intricate to "
                          "understand (and, consequently, error-prone).")

    for island in scr.get_islands():
        injectors = scr.injector.loc[scr.injector['bus'].isin(island)]
        if injectors.empty:
            raise ValueError("There's no injector connected to the island "
                             "comprising the buses " + str(list(island)))
        # A simple plausibility check of sufficient injection/load
        # (This can be helpful, e.g., in the simulation of contingencies,
        # where a line/transformer/converter fault can cause load or
        # generation buses to be islanded.)
        total_load = scr.bus.loc[island, 'load'].sum()
        if np.abs(total_load.real) < hynet_eps:
            _log.warning("There is no fixed load in the island comprising "
                         "the buses " + str(list(island)))
        if total_load.real > np.sum([x.p_max for x in injectors['cap']]):
            _log.warning("The active power load exceeds the injector "
                         "active power capacity in the island comprising "
                         "the buses " + str(list(island)))
        if total_load.real < np.sum([x.p_min for x in injectors['cap']]):
            _log.warning("The minimum active power injection exceeds the "
                         "active power load in the island comprising the "
                         "buses " + str(list(island)))

    # Objective checks

    for j, (cost_p, cost_q) in enumerate(zip(scr.injector['cost_p'].values,
                                             scr.injector['cost_q'].values)):
        if cost_p is not None:
            if not cost_p.is_convex():
                raise ValueError(("The P-cost function of injector {0} is "
                                  "nonconvex.").format(scr.injector.index[j]))
        if cost_q is not None:
            if not cost_q.is_convex():
                raise ValueError(("The Q-cost function of injector {0} is "
                                  "nonconvex.").format(scr.injector.index[j]))

    _log.debug("Scenario verification ({:.3f} sec.)".format(timer.total()))
